@@ -2,8 +2,9 @@
 
 import os
 from os.path import abspath
-from typing import Optional, Dict, List, Union
+from typing import Optional, Dict, List, Union, Any
 import pandas as pd
+from nerblackbox.modules.experiment_results import ExperimentResults
 
 try:
     from nerblackbox.modules.main import NerBlackBoxMain
@@ -63,15 +64,19 @@ class NerBlackBox:
         nerbb = NerBlackBoxMain("download")
         nerbb.main()
 
-    def get_experiment_results(self, experiment_name: str):
+    def get_experiment_results(self, experiment_name: str) -> List[ExperimentResults]:
         """get results for a single experiment.
 
         Args:
             experiment_name: e.g. "exp0"
+
+        Returns:
+            see ExperimentResults
         """
         kwargs = self._process_kwargs_optional()
         kwargs["usage"] = "api"
         kwargs["experiment_name"] = experiment_name
+        kwargs["from_config"] = True
 
         nerbb = NerBlackBoxMain("get_experiment_results", **kwargs)
         return nerbb.main()
@@ -92,22 +97,6 @@ class NerBlackBox:
         nerbb = NerBlackBoxMain("get_experiments", **kwargs)
         return nerbb.main()
 
-    def get_experiments_results(self, **kwargs_optional: Dict) -> Dict:
-        """get results from multiple experiments.
-
-        Args:
-            kwargs_optional: with optional key-value pairs \
-            {"ids": [tuple of int], "as_df": [bool]}
-
-        Returns:
-            experiments_results: w/ keys = "best_single_runs", "best_average_runs" \
-            & values = [pandas DataFrame] or [dict]
-        """
-        kwargs = self._process_kwargs_optional(kwargs_optional)
-        kwargs["usage"] = "api"
-        nerbb = NerBlackBoxMain("get_experiments_results", **kwargs)
-        return nerbb.main()
-
     def init(self):
         """initialize the data_dir directory.
         needs to be called exactly once before any other CLI/API commands of the package are executed.
@@ -133,17 +122,65 @@ class NerBlackBox:
         nerbb = NerBlackBoxMain("predict", **kwargs)
         return nerbb.main()
 
-    def run_experiment(self, experiment_name: str, **kwargs_optional: Dict):
+    def run_experiment(
+        self,
+        experiment_name: str,
+        from_config: bool = False,
+        model: Optional[str] = None,
+        dataset: Optional[str] = None,
+        from_preset: Optional[str] = "adaptive",
+        **kwargs_optional: Dict
+    ):
         """run a single experiment.
 
+           Note:
+
+           - from_config == True -> experiment config file is used, no other optional arguments will be used
+
+           - from_config == False -> experiment config file is created dynamically, optional arguments will be used
+
+               - model and dataset are mandatory.
+
+               - All other arguments relate to hyperparameters and are optional.
+                 If not specified, they are taken using the following hierarchy:
+
+                 1) optional argument
+
+                 2) from_preset (adaptive, original, stable),
+                    which specifies e.g. the hyperparameters "max_epochs", "early_stopping", "lr_schedule"
+
+                 3) default experiment configuration
+
+
         Args:
-            experiment_name: e.g. "exp0"
-            kwargs_optional: with optional key-value pairs \
-            {"run_name": [str], "device": [torch device], "fp16": [bool]}
+            experiment_name: e.g. 'exp0'
+            from_config: e.g. False
+            model: if experiment config file is to be created dynamically, e.g. 'bert-base-uncased'
+            dataset: if experiment config file is to be created dynamically, e.g. 'conll-2003'
+            from_preset: if experiment config file is to be created dynamically, e.g. 'adaptive'
+            kwargs_optional: with optional key-value pairs, e.g. \
+            {"multiple_runs": [int], "run_name": [str], "device": [torch device], "fp16": [bool]}
         """
 
         kwargs = self._process_kwargs_optional(kwargs_optional)
+
         kwargs["experiment_name"] = experiment_name
+        if model is not None:
+            kwargs["pretrained_model_name"] = model
+        if dataset is not None:
+            kwargs["dataset_name"] = dataset
+
+        kwargs["hparams"] = self._extract_hparams(kwargs)
+        kwargs["from_config"] = from_config
+        if not from_config:
+            kwargs["from_preset"] = from_preset
+
+        # get rid of keys in kwargs that are present in kwargs["hparams"]
+        for key in kwargs["hparams"].keys():
+            kwargs.pop(key)
+
+        if kwargs["hparams"] == {}:
+            kwargs["hparams"] = None
 
         nerbb = NerBlackBoxMain("run_experiment", **kwargs)
         nerbb.main()
@@ -164,10 +201,11 @@ class NerBlackBox:
         nerbb.main()
 
     def show_experiment_config(self, experiment_name: str):
-        """show a single experiment configuration in detail.
+        """show a single experiment configuration in detail
+           or an overview on all available experiment configurations.
 
         Args:
-            experiment_name: e.g. "exp0"
+            experiment_name: e.g. "exp0" or "all"
         """
         kwargs = self._process_kwargs_optional()
         kwargs["experiment_name"] = experiment_name
@@ -175,19 +213,41 @@ class NerBlackBox:
         nerbb = NerBlackBoxMain("show_experiment_config", **kwargs)
         nerbb.main()
 
-    def show_experiment_configs(self):
-        """show overview on all available experiment configurations."""
-        _ = self._process_kwargs_optional()
-
-        nerbb = NerBlackBoxMain("show_experiment_configs")
-        nerbb.main()
-
     ####################################################################################################################
     # HELPER
     ####################################################################################################################
     @staticmethod
-    def _process_kwargs_optional(_kwargs_optional: Optional[Dict] = None):
+    def _process_kwargs_optional(
+        _kwargs_optional: Optional[Dict[str, Any]] = None
+    ) -> Dict[str, Any]:
+        """
+        general helper function
+        filters out key-value pairs that have value = None
+
+        Args:
+            _kwargs_optional: e.g. {"a": 1, "b": None}
+
+        Returns:
+            _kwargs:          e.g. {"a": 1}
+        """
         if _kwargs_optional is None:
             return {}
         else:
             return {k: v for k, v in _kwargs_optional.items() if v is not None}
+
+    @staticmethod
+    def _extract_hparams(_kwargs: Dict[str, Any]) -> Dict[str, Any]:
+        """
+        Args:
+            _kwargs: e.g. {"a": 1, "run_name": "runA-1"}
+
+        Returns:
+            _hparams: e.g. {"a": 1}
+        """
+        # hparams
+        exclude_keys = ["experiment_name", "run_name", "device", "fp16"]
+        _hparams = {
+            _key: _kwargs[_key] for _key in _kwargs.keys() if _key not in exclude_keys
+        }
+
+        return _hparams
