@@ -24,6 +24,12 @@ from nerblackbox.modules.main import NerBlackBoxMain
 @click.option(
     "--verbose/--no-verbose", default=False, help="[bool] if flag=set_up_dataset"
 )
+@click.option(
+    "--from_config/--no-from_config",
+    default=True,
+    type=bool,
+    help="[bool] if flag=run_experiment",
+)
 @click.option("--run_name", default=None, type=str, help="[str] if flag=run_experiment")
 @click.option("--device", default=None, type=str, help="[str] if flag=run_experiment")
 @click.option("--fp16/--no-fp16", default=False, help="[bool] if flag=run_experiment")
@@ -122,16 +128,6 @@ def get_experiment_results(ctx, experiment_name: str):
     _run_nerblackbox_main(ctx.obj, kwargs)
 
 
-@nerbb.command(name="get_experiments_results")
-@click.pass_context
-def get_experiments_results(ctx):
-    """get results for multiple experiments."""
-    kwargs = {
-        "flag": "get_experiments_results",
-    }
-    _run_nerblackbox_main(ctx.obj, kwargs)
-
-
 @nerbb.command(name="init")
 @click.pass_context
 def init(ctx):
@@ -216,16 +212,6 @@ def show_experiment_config(ctx, experiment_name: str):
     _run_nerblackbox_main(ctx.obj, kwargs)
 
 
-@nerbb.command(name="show_experiment_configs")
-@click.pass_context
-def show_experiment_configs(ctx):
-    """show overview on all available experiment configurations."""
-    kwargs = {
-        "flag": "show_experiment_configs",
-    }
-    _run_nerblackbox_main(ctx.obj, kwargs)
-
-
 @nerbb.command(name="tensorboard")
 def tensorboard():
     """show detailed experiment results in tensorboard. (port = 6006)."""
