@@ -3,26 +3,24 @@ from os.path import join, isfile, isdir
 import glob
 import mlflow
 import shutil
-from argparse import Namespace
+from sys import exit
 from pkg_resources import Requirement
 from pkg_resources import resource_filename, resource_isdir
 import pandas as pd
 from mlflow.tracking import MlflowClient
+from mlflow.entities import Run
 
 from nerblackbox.modules.utils.env_variable import env_variable
-from nerblackbox.modules.utils.util_functions import epoch2checkpoint
-from nerblackbox.modules.utils.util_functions import (
-    get_run_name,
-    compute_mean_and_dmean,
-)
 from nerblackbox.modules.experiment_results import ExperimentResults
-from typing import Optional, Any, Tuple, Union, Dict, List
-from pandas import DataFrame
+from nerblackbox.modules.experiment_config.preset import get_preset
+from nerblackbox.modules.utils.parameters import DATASET, MODEL, SETTINGS, HPARAMS
+from typing import Optional, Tuple, Union, Dict, List
 
 DATASETS_DOWNLOAD = [
     "conll2003",
     "swedish_ner_corpus",
     "sic",
+    "swe_nerc",
 ]
 
 
@@ -36,8 +34,11 @@ class NerBlackBoxMain:
         val_fraction: float = 0.3,  # set_up_dataset
         verbose: bool = False,
         experiment_name: Optional[str] = None,
+        hparams: Optional[Dict[str, Union[str, int, bool]]] = None,  # run_experiment
+        from_preset: Optional[str] = None,  # run_experiment
+        from_config: bool = False,  # run_experiment
         run_name: Optional[str] = None,  # run_experiment
-        device: Any = "gpu",  # run_experiment
+        device: str = "gpu",  # run_experiment
         fp16: bool = False,  # run_experiment
         text_input: Optional[str] = None,  # predict
         ids: Tuple[str, ...] = (),  # get_experiments, get_experiments_results
@@ -47,13 +48,16 @@ class NerBlackBoxMain:
         """
         :param flag:            [str], e.g. 'analyze_data', 'set_up_dataset', 'run_experiment', ..
         :param usage:           [str] 'cli' or 'api'
-        :param dataset_name     [str] e.g. 'swedish_ner_corpus'
-        :param modify           [bool] if True: modify tags as specified in method modify_ner_tag_mapping()
-        :param val_fraction     [float] e.g. 0.3
-        :param verbose          [bool]
+        :param dataset_name:    [str] e.g. 'swedish_ner_corpus'
+        :param modify:          [bool] if True: modify tags as specified in method modify_ner_tag_mapping()
+        :param val_fraction:    [float] e.g. 0.3
+        :param verbose:         [bool]
         :param experiment_name: [str], e.g. 'exp0'
+        :param hparams:         [dict], e.g. {'multiple_runs': '2'} with hparams to use            [HIERARCHY:  I]
+        :param from_preset:     [str], e.g. 'adaptive' get experiment params & hparams from preset [HIERARCHY: II]
+        :param from_config:     [bool] if True, get experiment params & hparams from config file   [ALTERNATIVE]
         :param run_name:        [str or None], e.g. 'runA'
-        :param device:          [torch device]
+        :param device:          [str]
         :param fp16:            [bool]
         :param text_input:      [str], e.g. 'this is some text that needs to be annotated'
         :param ids:             [tuple of str], experiment_ids to include
@@ -71,6 +75,10 @@ class NerBlackBoxMain:
         self.val_fraction = val_fraction  # set_up_dataset
         self.verbose = verbose
         self.experiment_name = experiment_name
+        self.hparams: Optional[
+            Dict[str, Union[str, int, bool]]
+        ] = self._process_hparams(hparams, from_preset)
+        self.from_config: bool = from_config
         self.run_name = run_name  # run_experiment
         self.device = device  # run_experiment
         self.fp16 = fp16  # run_experiment
@@ -79,6 +87,38 @@ class NerBlackBoxMain:
         self.as_df = as_df  # get_experiments, get_experiments_results
         self.results = results  # clear_data
 
+        if self.flag == "run_experiment":
+            assert (self.hparams is None and self.from_config is True) or (
+                self.hparams is not None and self.from_config is False
+            ), (
+                f"ERROR! Need to specify "
+                f"EITHER hparams (currently {self.hparams}) "
+                f"with or without from_preset (currently {from_preset}) "
+                f"OR from_config (currently {self.from_config})."
+            )
+
+            if self.from_config:
+                path_experiment_config = join(
+                    env_variable("DIR_EXPERIMENT_CONFIGS"),
+                    f"{self.experiment_name}.ini",
+                )
+                if not isfile(path_experiment_config):
+                    self._exit_gracefully(
+                        f"experiment_config = {path_experiment_config} does not exist."
+                    )
+            else:
+                assert (
+                    self.hparams is not None
+                ), f"ERROR! self.hparams is None but needs to be specified if dynamic arguments are used."
+                for field in ["pretrained_model_name", "dataset_name"]:
+                    if field not in self.hparams.keys():
+                        field_displayed = (
+                            "model" if field == "pretrained_model_name" else "dataset"
+                        )
+                        self._exit_gracefully(
+                            f"{field_displayed} is not specified but mandatory if dynamic arguments are used."
+                        )
+
         data_dir = env_variable("DATA_DIR")
         if os.path.isdir(data_dir):
             self._set_client_and_get_experiments()
@@ -88,6 +128,12 @@ class NerBlackBoxMain:
             self.experiment_id2name = None
             self.experiment_name2id = None
 
+    @staticmethod
+    def _exit_gracefully(message: str) -> None:
+        print(message)
+        print("stopped.")
+        exit(0)
+
     ####################################################################################################################
     # MAIN #############################################################################################################
     ####################################################################################################################
@@ -121,12 +167,6 @@ class NerBlackBoxMain:
             self._assert_flag_arg("dataset_name")
             self.set_up_dataset(self.dataset_name)
 
-        ################################################################################################################
-        # show_experiment_configs
-        ################################################################################################################
-        elif self.flag == "show_experiment_configs":
-            self.show_experiment_configs()
-
         ################################################################################################################
         # show_experiment_config
         ################################################################################################################
@@ -145,7 +185,6 @@ class NerBlackBoxMain:
         # get_experiment_results
         ################################################################################################################
         elif self.flag == "get_experiment_results":
-            self._assert_flag_arg("experiment_name")
             self._assert_usage()
             return self.get_experiment_results()
 
@@ -156,13 +195,6 @@ class NerBlackBoxMain:
             self._assert_usage()
             return self.get_experiments()
 
-        ################################################################################################################
-        # get_experiments_best_runs
-        ################################################################################################################
-        elif self.flag == "get_experiments_results":
-            self._assert_usage()
-            return self.get_experiments_results()
-
         ################################################################################################################
         # predict
         ################################################################################################################
@@ -253,10 +285,13 @@ class NerBlackBoxMain:
                 else:
                     print("Please enter either y or n")
 
-    def get_experiment_results(self) -> Optional[ExperimentResults]:
+    def get_experiment_results(self) -> Optional[List[ExperimentResults]]:
         """
-        :used attr: experiment_name [str], e.g. 'exp0'
-        :return: experiment_results [ExperimentResults]
+        Used Attr:
+            experiment_name: e.g. 'all', 'exp0'
+
+        Returns:
+            experiment_results_list: returned for API, not CLI
         """
         from nerblackbox.modules.ner_training.ner_model_predict import (
             NerModelPredict,
@@ -265,32 +300,81 @@ class NerBlackBoxMain:
         assert (
             self.experiment_id2name is not None
         ), f"ERROR! self.experiment_id2name is None."
-        if self.experiment_name not in self.experiment_name2id.keys():
+        if self.experiment_name == "all":
+            # all experiments
+            assert self.ids is not None, f"ERROR! self.ids is None."
+            experiments_id2name = self._filter_experiments_by_ids(self.ids)
+        elif self.experiment_name in self.experiment_name2id.keys():
+            # single experiment
+            experiments_id2name = {
+                self.experiment_name2id[self.experiment_name]: self.experiment_name
+            }
+        else:
             print(f"no experiment with experiment_name = {self.experiment_name} found")
             print(f"experiments that were found:")
             print(self.experiment_name2id)
-            return ExperimentResults()
+            return [ExperimentResults()]
 
-        experiment_id = self.experiment_name2id[self.experiment_name]
-        experiment_results = self._get_single_experiment_results(experiment_id)
+        # get ExperimentResults
+        experiment_results_list: List[ExperimentResults] = list()
+        for _id in sorted(list(experiments_id2name.keys())):
+            experiment_results_list.append(self._get_single_experiment_results(_id))
 
+        # return
         if self.usage == "cli":
-            print("### single runs ###")
-            print(experiment_results.single_runs)
-            print()
-            print("### average runs ###")
-            print(experiment_results.average_runs)
+            if self.experiment_name != "all":
+                # single experiment
+                print("### single runs ###")
+                print(experiment_results_list[0].single_runs.T)
+                print()
+                print("### average runs ###")
+                print(experiment_results_list[0].average_runs.T)
+            else:
+                # all experiments
+                best_single_runs_overview = [
+                    experiment_results.best_single_run
+                    for experiment_results in experiment_results_list
+                ]
+                for best_single_runs_elem in best_single_runs_overview:
+                    if "checkpoint" in best_single_runs_elem.keys():
+                        best_single_runs_elem.pop("checkpoint")
+
+                df_single = (
+                    pd.DataFrame(best_single_runs_overview)
+                    if self.as_df
+                    else best_single_runs_overview
+                )
+                best_average_runs_overview = [
+                    experiment_results.best_average_run
+                    for experiment_results in experiment_results_list
+                ]
+                df_average = (
+                    pd.DataFrame(best_average_runs_overview)
+                    if self.as_df
+                    else best_average_runs_overview
+                )
+                print("### best single runs ###")
+                print(df_single.T)
+                print()
+                print("### best average runs ###")
+                print(df_average.T)
             return None
         else:
-            if experiment_results.best_single_run["checkpoint"] is not None:
-                best_model = NerModelPredict.load_from_checkpoint(
-                    experiment_results.best_single_run["checkpoint"]
-                )
-                experiment_results._set_best_model(best_model)
+            if self.experiment_name != "all":
+                # single experiment
+                if (
+                    "checkpoint" in experiment_results_list[0].best_single_run.keys()
+                    and experiment_results_list[0].best_single_run["checkpoint"]
+                    is not None
+                ):
+                    best_model = NerModelPredict.load_from_checkpoint(
+                        experiment_results_list[0].best_single_run["checkpoint"]
+                    )
+                    experiment_results_list[0].set_best_model(best_model)
 
-            return experiment_results
+            return experiment_results_list
 
-    def get_experiments(self) -> Optional[Union[DataFrame, Dict]]:
+    def get_experiments(self) -> Optional[Union[pd.DataFrame, Dict]]:
         r"""
         :used attr: ids    [tuple] of [str], e.g. ('4', '5')
         :used attr: as_df  [bool] if True, return [pandas DataFrame], else return [dict]
@@ -314,53 +398,6 @@ class NerBlackBoxMain:
         else:
             return df
 
-    def get_experiments_results(self) -> Optional[Namespace]:
-        r"""
-        :used attr: ids     [tuple] of [str], e.g. ('4', '5')
-        :used attr: as_df   [bool] if True, return [pandas DataFrame], else return [dict]
-        :used attr: verbose [bool]
-        :return: experiments_results: [Namespace] w/ attributes = 'best_single_runs', 'best_average_runs'
-                                                   & values = [pandas DataFrame] or [dict]
-        """
-        assert self.ids is not None, f"ERROR! self.ids is None."
-        experiments_filtered = self._filter_experiments_by_ids(self.ids)
-
-        best_single_runs_overview = list()
-        best_average_runs_overview = list()
-        for _id in sorted(list(experiments_filtered.keys())):
-            experiment_results = self._get_single_experiment_results(
-                _id,
-            )
-            if experiment_results.best_single_run:
-                best_single_runs_overview.append(experiment_results.best_single_run)
-            if experiment_results.best_average_run:
-                best_average_runs_overview.append(experiment_results.best_average_run)
-
-        df_single = (
-            pd.DataFrame(best_single_runs_overview)
-            if self.as_df
-            else best_single_runs_overview
-        )
-        df_average = (
-            pd.DataFrame(best_average_runs_overview)
-            if self.as_df
-            else best_average_runs_overview
-        )
-        if self.usage == "cli":
-            print("### best single runs ###")
-            print(df_single)
-            print()
-            print("### best average runs ###")
-            print(df_average)
-            return None
-        else:
-            return Namespace(
-                **{
-                    "best_single_runs": df_single,
-                    "best_average_runs": df_average,
-                }
-            )
-
     def predict(self) -> Optional[List[List[Dict[str, str]]]]:
         """
         :used attr: experiment_name [str], e.g. 'exp1'
@@ -372,14 +409,14 @@ class NerBlackBoxMain:
             "get_experiment_results", experiment_name=self.experiment_name, usage="api"
         )
         experiment_results = nerbb.main()
-        predictions = experiment_results.best_model.predict(self.text_input)
+        predictions = experiment_results[0].best_model.predict(self.text_input)
         if self.usage == "cli":
             print(predictions[0])
             return None
         else:
             return predictions
 
-    def predict_proba(self) -> Optional[List[Namespace]]:
+    def predict_proba(self) -> Optional[List[Dict[str, Union[str, Dict]]]]:
         """
         :used attr: experiment_name [str], e.g. 'exp1'
         :used attr: text_input      [str], e.g. 'this is some text that needs to be annotated'
@@ -392,25 +429,31 @@ class NerBlackBoxMain:
         experiment_results = nerbb.main()
         predictions = experiment_results.best_model.predict_proba(self.text_input)
         if self.usage == "cli":
-            print(predictions[0].external)
+            print(predictions[0])
             return None
         else:
-            return predictions
+            return predictions[0]
 
     def run_experiment(self) -> None:
         """
         :used attr: experiment_name [str],         e.g. 'exp1'
+        :used attr: hparams         [dict], e.g. {'multiple_runs': '2'} with hparams to use           [HIERARCHY:  I]
+        :used attr: from_config     [bool] if True, get experiment params & hparams from config file  [ALTERNATIVE]
         :used attr: run_name        [str] or None, e.g. 'runA'
-        :used attr: device          [torch device]
+        :used attr: device          [str]
         :used attr: fp16            [bool]
         """
         _parameters = {
             "experiment_name": self.experiment_name,
+            "from_config": int(self.from_config),
             "run_name": self.run_name if self.run_name else "",
             "device": self.device,
             "fp16": int(self.fp16),
         }
 
+        if self.from_config is False:
+            self._write_config_file()
+
         mlflow.projects.run(
             uri=resource_filename(Requirement.parse("nerblackbox"), "nerblackbox"),
             entry_point="run_experiment",
@@ -449,35 +492,101 @@ class NerBlackBoxMain:
         """
         print experiment config
         -----------------------
-        :used attr: experiment_name: [str], e.g. 'exp0'
+        :used attr: experiment_name: [str], e.g. 'exp0' or 'all'
         """
         from nerblackbox.modules.utils.env_variable import env_variable
 
-        path_experiment_config = join(
-            env_variable("DIR_EXPERIMENT_CONFIGS"), f"{self.experiment_name}.ini"
-        )
-        with open(path_experiment_config, "r") as file:
-            lines = file.read()
+        if self.experiment_name != "all":
+            path_experiment_config = join(
+                env_variable("DIR_EXPERIMENT_CONFIGS"), f"{self.experiment_name}.ini"
+            )
+            if isfile(path_experiment_config):
+                with open(path_experiment_config, "r") as file:
+                    lines = file.read()
 
-        print(f"> experiment_config = {path_experiment_config}")
-        print()
-        print(lines)
-
-    @staticmethod
-    def show_experiment_configs() -> None:
-        experiment_configs = glob.glob(
-            join(env_variable("DIR_EXPERIMENT_CONFIGS"), "*.ini")
-        )
-        experiment_configs = [
-            elem.split("/")[-1].strip(".ini") for elem in experiment_configs
-        ]
-        experiment_configs = [elem for elem in experiment_configs if elem != "default"]
-        for experiment_config in experiment_configs:
-            print(experiment_config)
+                print(f"> experiment_config = {path_experiment_config}")
+                print()
+                print(lines)
+            else:
+                print(f"> experiment_config = {path_experiment_config} does not exist.")
+        else:
+            experiment_configs = glob.glob(
+                join(env_variable("DIR_EXPERIMENT_CONFIGS"), "*.ini")
+            )
+            experiment_configs = [
+                elem.split("/")[-1].strip(".ini") for elem in experiment_configs
+            ]
+            experiment_configs = [
+                elem for elem in experiment_configs if elem != "default"
+            ]
+            for experiment_config in experiment_configs:
+                print(experiment_config)
 
     ####################################################################################################################
     # HELPER ###########################################################################################################
     ####################################################################################################################
+    @staticmethod
+    def _process_hparams(
+        hparams: Optional[Dict[str, Union[str, int, bool]]], from_preset: Optional[str]
+    ) -> Optional[Dict[str, Union[str, int, bool]]]:
+        """
+        Args:
+            hparams:         [dict], e.g. {'multiple_runs': '2'} with hparams to use            [HIERARCHY:  I]
+            from_preset:     [str], e.g. 'adaptive' get experiment params & hparams from preset [HIERARCHY: II]
+
+        Returns:
+            _hparams:        [dict], e.g. {'multiple_runs': '2'} with hparams to use
+        """
+        _hparams = get_preset(from_preset)
+        if _hparams is None:
+            _hparams = hparams
+        elif hparams is not None:
+            _hparams.update(**hparams)
+        return _hparams
+
+    def _write_config_file(self) -> None:
+        """
+        write config file based on self.hparams
+        """
+        # assert that config file does not exist
+        config_path = join(
+            env_variable("DIR_EXPERIMENT_CONFIGS"), f"{self.experiment_name}.ini"
+        )
+        assert (
+            isfile(config_path) is False
+        ), f"ERROR! experiment config file {config_path} already exists!"
+
+        # write config file: helper functions
+        def _write(_str: str):
+            f.write(_str + "\n")
+
+        def _write_key_value(_key: str):
+            assert (
+                self.hparams is not None
+            ), f"ERROR! self.hparams is None - _write_key_value() failed."
+            if _key in self.hparams.keys():
+                f.write(f"{_key} = {self.hparams[_key]}\n")
+
+        # write config file
+        with open(config_path, "w") as f:
+            _write("[dataset]")
+            for key in DATASET.keys():
+                _write_key_value(key)
+
+            _write("\n[model]")
+            for key in MODEL.keys():
+                _write_key_value(key)
+
+            _write("\n[settings]")
+            for key in SETTINGS.keys():
+                _write_key_value(key)
+
+            _write("\n[hparams]")
+            for key in HPARAMS.keys():
+                _write_key_value(key)
+
+            _write("\n[runA]")
+
     @staticmethod
     def _create_data_directory() -> None:
         if resource_isdir(Requirement.parse("nerblackbox"), "nerblackbox/modules/data"):
@@ -539,91 +648,19 @@ class NerBlackBoxMain:
             self.experiment_id2name is not None
         ), f"ERROR! self.experiment_id2name is None."
 
-        experiment_name = self.experiment_id2name[experiment_id]
-        runs = self.client.search_runs(experiment_id)
+        experiment_name: str = self.experiment_id2name[experiment_id]
+        runs: List[Run] = self.client.search_runs(experiment_id)
 
-        _experiment, _single_runs, _average_runs = self._parse_and_create_dataframe(
+        return ExperimentResults.from_mlflow_runs(
             runs,
-        )
-
-        # best run
-        if _experiment is not None and _single_runs is not None:
-            _df_best_single_run = _single_runs.iloc[0, :]
-            best_single_run_id = _df_best_single_run[("info", "run_id")]
-            best_single_run_name_nr = _df_best_single_run[("info", "run_name_nr")]
-            best_single_run_epoch_best = _df_best_single_run[
-                ("metrics", "epoch_best".upper())
-            ]
-            best_single_run_epoch_best_val_entity_fil_f1_micro = _df_best_single_run[
-                ("metrics", "epoch_best_val_entity_fil_f1_micro".upper())
-            ]
-            best_single_run_epoch_best_test_entity_fil_f1_micro = _df_best_single_run[
-                ("metrics", "epoch_best_test_entity_fil_f1_micro".upper())
-            ]
-
-            checkpoint = join(
-                env_variable("DIR_CHECKPOINTS"),
-                experiment_name,
-                best_single_run_name_nr,
-                epoch2checkpoint(best_single_run_epoch_best),
-            )
-
-            _best_single_run = {
-                "experiment_id": experiment_id,
-                "experiment_name": experiment_name,
-                "run_id": best_single_run_id,
-                "run_name_nr": best_single_run_name_nr,
-                "epoch_best_val_entity_fil_f1_micro".upper(): best_single_run_epoch_best_val_entity_fil_f1_micro,
-                "epoch_best_test_entity_fil_f1_micro".upper(): best_single_run_epoch_best_test_entity_fil_f1_micro,
-                "checkpoint": checkpoint if isfile(checkpoint) else None,
-            }
-        else:
-            _best_single_run = dict()
-
-        # best run average
-        if _experiment is not None and _average_runs is not None:
-            _df_best_average_run = _average_runs.iloc[0, :]
-            best_average_run_name = _df_best_average_run[("info", "run_name")]
-            best_average_run_epoch_best_val_entity_fil_f1_micro = _df_best_average_run[
-                ("metrics", "epoch_best_val_entity_fil_f1_micro".upper())
-            ]
-            d_best_average_run_epoch_best_val_entity_fil_f1_micro = (
-                _df_best_average_run[
-                    ("metrics", "d_epoch_best_val_entity_fil_f1_micro".upper())
-                ]
-            )
-            best_average_run_epoch_best_test_entity_fil_f1_micro = _df_best_average_run[
-                ("metrics", "epoch_best_test_entity_fil_f1_micro".upper())
-            ]
-            d_best_average_run_epoch_best_test_entity_fil_f1_micro = (
-                _df_best_average_run[
-                    ("metrics", "d_epoch_best_test_entity_fil_f1_micro".upper())
-                ]
-            )
-
-            _best_average_run = {
-                "experiment_id": experiment_id,
-                "experiment_name": experiment_name,
-                "run_name": best_average_run_name,
-                "epoch_best_val_entity_fil_f1_micro".upper(): best_average_run_epoch_best_val_entity_fil_f1_micro,
-                "d_epoch_best_val_entity_fil_f1_micro".upper(): d_best_average_run_epoch_best_val_entity_fil_f1_micro,
-                "epoch_best_test_entity_fil_f1_micro".upper(): best_average_run_epoch_best_test_entity_fil_f1_micro,
-                "d_epoch_best_test_entity_fil_f1_micro".upper(): d_best_average_run_epoch_best_test_entity_fil_f1_micro,
-            }
-        else:
-            _best_average_run = dict()
-        return ExperimentResults(
-            _experiment,
-            _single_runs,
-            _average_runs,
-            _best_single_run,
-            _best_average_run,
+            experiment_id,
+            experiment_name,
         )
 
     ####################################################################################################################
     # HELPER: ALL EXPERIMENTS
     ####################################################################################################################
-    def _filter_experiments_by_ids(self, _ids: Tuple[str, ...]) -> Dict:
+    def _filter_experiments_by_ids(self, _ids: Tuple[str, ...]) -> Dict[str, str]:
         r"""
         get _experiments_id2name [dict] with _ids as keys
         -------------------------------------------------
@@ -642,179 +679,8 @@ class NerBlackBoxMain:
             }
         return _experiments_id2name
 
-    @staticmethod
-    def _parse_and_create_dataframe(
-        _runs: List,
-    ) -> Tuple[DataFrame, DataFrame, DataFrame]:
-        r"""
-        turn mlflow Run objects (= search_runs() results) into data frames
-        ------------------------------------------------------------------
-        :param _runs:   [list] of [mlflow.entities.Run objects]
-        :return: _experiment:   [pandas DataFrame] overview on experiment parameters
-        :return: _single_runs:  [pandas DataFrame] overview on single  run parameters & results
-        :return: _average_runs: [pandas DataFrame] overview on average run parameters & results
-        """
-        fields_metrics = [
-            "epoch_best".upper(),
-            "epoch_stopped".upper(),
-            "epoch_best_val_entity_fil_f1_micro".upper(),
-            "epoch_best_test_entity_fil_f1_micro".upper(),
-        ]
-
-        ###########################################
-        # parameters_experiment & parameters_runs
-        ###########################################
-        parameters_runs: Dict[Tuple, Any] = dict()
-        parameters_experiment: Dict[Any, Any] = dict()
-        for i in range(len(_runs)):
-            if len(_runs[i].data.metrics) == 0:  # experiment
-                parameters_experiment = {
-                    k: [v] for k, v in _runs[i].data.params.items()
-                }
-            else:  # run
-                if ("info", "run_id") not in parameters_runs.keys():
-                    parameters_runs[("info", "run_id")] = [_runs[i].info.run_id]
-                else:
-                    parameters_runs[("info", "run_id")].append(_runs[i].info.run_id)
-
-                if ("info", "run_name_nr") not in parameters_runs.keys():
-                    parameters_runs[("info", "run_name_nr")] = [
-                        _runs[i].data.tags["mlflow.runName"]
-                    ]
-                else:
-                    parameters_runs[("info", "run_name_nr")].append(
-                        _runs[i].data.tags["mlflow.runName"]
-                    )
-
-                for k, v in _runs[i].data.params.items():
-                    if ("params", k) not in parameters_runs.keys():
-                        parameters_runs[("params", k)] = [v]
-                    else:
-                        parameters_runs[("params", k)].append(v)
-
-                for k in fields_metrics:
-                    if ("metrics", k) not in parameters_runs.keys():
-                        try:
-                            parameters_runs[("metrics", k)] = [_runs[i].data.metrics[k]]
-                        except:
-                            parameters_runs[("metrics", k)] = [-1]
-                    else:
-                        try:
-                            parameters_runs[("metrics", k)].append(
-                                _runs[i].data.metrics[k]
-                            )
-                        except:
-                            parameters_runs[("metrics", k)] = [-1]
-
-        _experiment = pd.DataFrame(parameters_experiment, index=["experiment"]).T
-        for k in ["epoch_best".upper(), "epoch_stopped".upper()]:
-            try:
-                parameters_runs[("metrics", k)] = [
-                    int(elem) for elem in parameters_runs[("metrics", k)]
-                ]
-            except:
-                parameters_runs[("metrics", k)] = [-1]
-
-        ###########################################
-        # parameters_runs_average
-        ###########################################
-        def average(_parameters_runs):
-            _parameters_runs_average = {("info", "run_name"): list()}
-            _parameters_runs_average.update(
-                {
-                    k: list()
-                    for k in _parameters_runs.keys()
-                    if k[0] not in ["info", "metrics"]
-                }
-            )
-            _parameters_runs_average.update(
-                {
-                    k: list()
-                    for k in [
-                        ("metrics", "epoch_best_val_entity_fil_f1_micro".upper()),
-                        ("metrics", "d_epoch_best_val_entity_fil_f1_micro".upper()),
-                        ("metrics", "epoch_best_test_entity_fil_f1_micro".upper()),
-                        ("metrics", "d_epoch_best_test_entity_fil_f1_micro".upper()),
-                    ]
-                }
-            )
-
-            runs_name_nr = _parameters_runs[("info", "run_name_nr")]
-            nr_runs = len(runs_name_nr)
-            runs_name = list(
-                set([get_run_name(run_name_nr) for run_name_nr in runs_name_nr])
-            )
-            indices = {
-                run_name: [
-                    idx
-                    for idx in range(nr_runs)
-                    if get_run_name(runs_name_nr[idx]) == run_name
-                ]
-                for run_name in runs_name
-            }
-
-            def get_mean_and_dmean(_parameters_runs, phase):
-                values = [
-                    _parameters_runs[
-                        ("metrics", f"epoch_best_{phase}_entity_fil_f1_micro".upper())
-                    ][idx]
-                    for idx in indices[run_name]
-                ]
-                return compute_mean_and_dmean(values)
-
-            #######################
-            # loop over runs_name
-            #######################
-            for run_name in runs_name:
-                # add ('info', 'run_name')
-                _parameters_runs_average[("info", "run_name")].append(run_name)
-
-                # add ('params', *)
-                random_index = indices[run_name][0]
-                keys_kept = [k for k in _parameters_runs.keys() if k[0] == "params"]
-                for key in keys_kept:
-                    _parameters_runs_average[key].append(
-                        _parameters_runs[key][random_index]
-                    )
-
-                # add ('metrics', *)
-                val_mean, val_dmean = get_mean_and_dmean(_parameters_runs, "val")
-                test_mean, test_dmean = get_mean_and_dmean(_parameters_runs, "test")
-                metrics = {
-                    "epoch_best_val_entity_fil_f1_micro".upper(): val_mean,
-                    "d_epoch_best_val_entity_fil_f1_micro".upper(): val_dmean,
-                    "epoch_best_test_entity_fil_f1_micro".upper(): test_mean,
-                    "d_epoch_best_test_entity_fil_f1_micro".upper(): test_dmean,
-                }
-                for k in metrics.keys():
-                    key = ("metrics", k)
-                    _parameters_runs_average[key].append(metrics[k])
-
-            return _parameters_runs_average
-
-        ###########################################
-        # sort & return
-        ###########################################
-        by = ("metrics", "epoch_best_val_entity_fil_f1_micro".upper())
-        try:
-            _single_runs = pd.DataFrame(parameters_runs).sort_values(
-                by=by, ascending=False
-            )
-        except:
-            _single_runs = None
-
-        try:
-            parameters_runs_average = average(parameters_runs)
-            _average_runs = pd.DataFrame(parameters_runs_average).sort_values(
-                by=by, ascending=False
-            )
-        except:
-            _average_runs = None
-
-        return _experiment, _single_runs, _average_runs
-
     ####################################################################################################################
-    # HELPER: ADDITONAL
+    # HELPER: ADDITIONAL
     ####################################################################################################################
     @staticmethod
     def _assert_flag(flag: str) -> None:
@@ -833,5 +699,5 @@ class NerBlackBoxMain:
         assert self.usage in ["cli", "api"], "missing usage"
 
     @staticmethod
-    def show_as_df(_dict: Dict) -> DataFrame:
+    def show_as_df(_dict: Dict) -> pd.DataFrame:
         return pd.DataFrame(_dict)
