@@ -1,6 +1,6 @@
 from dataclasses import dataclass
 from dataclasses import asdict
-from typing import List, Optional, Tuple, Callable
+from typing import List, Tuple, Callable
 
 import numpy as np
 from sklearn.metrics import accuracy_score as accuracy_sklearn
@@ -14,6 +14,9 @@ import warnings
 from seqeval.metrics import precision_score as precision_seqeval
 from seqeval.metrics import recall_score as recall_seqeval
 from seqeval.metrics import f1_score as f1_seqeval
+from seqeval.scheme import IOB2, BILOU
+
+from nerblackbox.modules.ner_training.annotation_tags.tags import Tags
 
 
 class NerMetrics:
@@ -26,27 +29,36 @@ class NerMetrics:
         self,
         true_flat,
         pred_flat,
-        tag_list=None,
-        tag_index=None,
-        level="token",
-        plain_tags=False,
+        level,
+        scheme,
+        classes=None,
+        class_index=None,
         verbose=False,
     ):
         """
-        :param true_flat: [np array] of shape [batch_size * seq_length]
-        :param pred_flat: [np array] of shape [batch_size * seq_length]
-        :param tag_list:  [optional, list] of [str] labels to take into account for metrics -> if level = 'token'
-        :param tag_index: [optional, int]            index to take into account for metrics -> if level = 'entity'
-        :param level:     [optional, str] 'token' or 'entity'
-        :param verbose:   [optional, bool] if True, show verbose output
+        :param true_flat:   [np array] of shape [batch_size * seq_length]
+        :param pred_flat:   [np array] of shape [batch_size * seq_length]
+        :param level:       [str] 'token' or 'entity'
+        :param scheme:      [str] e.g. 'plain', 'bio'
+        :param classes:     [optional, list] of [str] labels to take into account for metrics -> if level = 'token'
+        :param class_index: [optional, int]            index to take into account for metrics -> if level = 'entity'
+        :param verbose:     [optional, bool] if True, show verbose output
         """
-        self.true_flat = true_flat
-        self.pred_flat = pred_flat
-        self.tag_list = tag_list
-        self.tag_index = tag_index
+        self.true_flat = true_flat  # token -> plain. entity -> plain, bio, bilou
+        self.pred_flat = pred_flat  # token -> plain. entity -> plain, bio, bilou
+        self.scheme = scheme  # token -> plain. entity -> plain, bio, bilou
+        self.classes = classes
+        self.class_index = class_index
         self.level = level
         self.verbose = verbose
 
+        if self.scheme == "bilou":
+            self.scheme_entity = "bilou"
+            self.scheme_entity_seqeval = BILOU
+        else:  # plain, bio
+            self.scheme_entity = "bio"
+            self.scheme_entity_seqeval = IOB2
+
         self.results = Results()
         self.failure_value = -1
 
@@ -55,12 +67,21 @@ class NerMetrics:
             "entity",
         ], f"ERROR! level = {self.level} unknown."
         if self.level == "entity":
-            self.true_flat_bio: List[str] = convert2bio(
-                self.true_flat, convert_to_bio=plain_tags
-            )
-            self.pred_flat_bio: List[str] = convert2bio(
-                self.pred_flat, convert_to_bio=plain_tags
-            )
+            self.true_flat_bio: List[str] = Tags(self.true_flat,).convert_scheme(
+                source_scheme=self.scheme, target_scheme=self.scheme_entity
+            )  # entity -> bio, bilou
+
+            self.pred_flat_bio: List[str] = Tags(self.pred_flat).convert_scheme(
+                source_scheme=self.scheme, target_scheme=self.scheme_entity
+            )  # entity -> bio, bilou
+
+            # ASR
+            self.pred_flat_bio_corrected: List[str]
+            self.pred_flat_bio_corrected, self.results.asr_abidance = Tags(
+                self.pred_flat_bio
+            ).restore_annotation_scheme_consistency(
+                scheme=self.scheme_entity
+            )  # entity -> bio, bilou
 
     def results_as_dict(self):
         return asdict(self.results)
@@ -86,6 +107,14 @@ class NerMetrics:
             if "f1" in _metrics:
                 self.f1_score()
 
+        if (
+            "asr_abidance" in _metrics
+            or "asr_precision" in _metrics
+            or "asr_recall" in _metrics
+            or "asr_f1" in _metrics
+        ):
+            self.compute_asr_results()
+
         warnings.resetwarnings()
 
     def accuracy(self):
@@ -119,7 +148,7 @@ class NerMetrics:
                 evaluation_function=precision_seqeval
             )
             self.results.precision_macro = self._entity_evaluation_macro(
-                evaluation_function=precision_seqeval, restrict_macro=True
+                evaluation_function=precision_seqeval,
             )
 
     def recall(self):
@@ -142,7 +171,7 @@ class NerMetrics:
                 evaluation_function=recall_seqeval
             )
             self.results.recall_macro = self._entity_evaluation_macro(
-                evaluation_function=recall_seqeval, restrict_macro=True
+                evaluation_function=recall_seqeval
             )
 
     def f1_score(self):
@@ -162,9 +191,43 @@ class NerMetrics:
             )
         elif self.level == "entity":
             self.results.f1_micro, self.results.f1_macro = self._entity_evaluation_f1(
-                evaluation_function=f1_seqeval, restrict_macro=True
+                evaluation_function=f1_seqeval,
             )
 
+    def compute_asr_results(self):
+        """
+        computes
+        - self.results.asr_precision_micro
+        - self.results.asr_recall_micro
+        - self.results.asr_f1_micro
+        """
+
+        def _entity_evaluation_micro_asr(evaluation_function: Callable) -> float:
+            """helper function"""
+            try:
+                metric = evaluation_function(
+                    [self.true_flat_bio],
+                    [self.pred_flat_bio_corrected],  # corrected !!!
+                    average="micro",
+                    mode="strict",
+                    scheme=self.scheme_entity_seqeval,
+                )
+            except UndefinedMetricWarning as e:
+                if self.verbose:
+                    print(e)
+                metric = self.failure_value
+            return metric
+
+        self.results.asr_precision_micro = _entity_evaluation_micro_asr(
+            evaluation_function=precision_seqeval
+        )
+        self.results.asr_recall_micro = _entity_evaluation_micro_asr(
+            evaluation_function=recall_seqeval
+        )
+        self.results.asr_f1_micro = _entity_evaluation_micro_asr(
+            evaluation_function=f1_seqeval
+        )
+
     def _token_evaluation(self, evaluation_function: Callable, average: str) -> float:
         """
         compute precision/recall/f1 on token level
@@ -183,26 +246,46 @@ class NerMetrics:
         ], f"evaluation function = {evaluation_function} unknown / not allowed."
         assert average in ["micro", "macro"], f"average = {average} unknown."
 
-        try:
+        if self.classes is None or len(self.classes) > 1:  # "all" / "fil"
             if evaluation_function != prf_sklearn:
                 metric = evaluation_function(
                     self.true_flat,
                     self.pred_flat,
-                    labels=self.tag_list,
+                    labels=self.classes,
                     average=average,
+                    zero_division=0,
                 )
             else:
                 _, _, metric, _ = prf_sklearn(
                     self.true_flat,
                     self.pred_flat,
-                    labels=self.tag_list,
+                    labels=self.classes,
                     average=average,
-                    warn_for=("precision", "recall", "f-score"),
+                    zero_division=0,
                 )
-        except UndefinedMetricWarning as e:
-            if self.verbose:
-                print(e)
-            metric = self.failure_value
+        else:
+            try:
+                if evaluation_function != prf_sklearn:
+                    metric = evaluation_function(
+                        self.true_flat,
+                        self.pred_flat,
+                        labels=self.classes,
+                        average=average,
+                        zero_division="warn",
+                    )
+                else:
+                    _, _, metric, _ = prf_sklearn(
+                        self.true_flat,
+                        self.pred_flat,
+                        labels=self.classes,
+                        average=average,
+                        warn_for=("precision", "recall", "f-score"),
+                        zero_division="warn",
+                    )
+            except UndefinedMetricWarning as e:
+                if self.verbose:
+                    print(e)
+                metric = self.failure_value
 
         return metric
 
@@ -221,10 +304,14 @@ class NerMetrics:
             recall_seqeval,
         ], f"evaluation function = {evaluation_function} unknown / not allowed."
 
-        if self.tag_index is None:  # "fil"
+        if self.class_index is None:  # "fil"
             try:
                 metric = evaluation_function(
-                    [self.true_flat_bio], [self.pred_flat_bio], average="micro"
+                    [self.true_flat_bio],
+                    [self.pred_flat_bio],
+                    average="micro",
+                    mode="strict",
+                    scheme=self.scheme_entity_seqeval,
                 )
             except UndefinedMetricWarning as e:
                 if self.verbose:
@@ -235,17 +322,21 @@ class NerMetrics:
                 metric = evaluation_function(
                     [self.true_flat_bio],
                     [self.pred_flat_bio],
+                    mode="strict",
+                    scheme=self.scheme_entity_seqeval,
                     average=None,
                     zero_division="warn",
-                )[self.tag_index]
+                )[self.class_index]
             except UndefinedMetricWarning:
                 try:
                     metric = evaluation_function(
                         [self.true_flat_bio],
                         [self.pred_flat_bio],
+                        mode="strict",
+                        scheme=self.scheme_entity_seqeval,
                         average=None,
                         zero_division=0,
-                    )[self.tag_index]
+                    )[self.class_index]
                 except IndexError:
                     metric = self.failure_value
 
@@ -253,9 +344,11 @@ class NerMetrics:
                     metric = evaluation_function(
                         [self.true_flat_bio],
                         [self.pred_flat_bio],
+                        mode="strict",
+                        scheme=self.scheme_entity_seqeval,
                         average=None,
                         zero_division=1,
-                    )[self.tag_index]
+                    )[self.class_index]
                     if metric == 1:
                         metric = self.failure_value
             except IndexError:
@@ -270,18 +363,30 @@ class NerMetrics:
             results.numberofclasses_macro: number of well-defined classes in terms of precision, recall, f1
         """
 
-        def _get_index_list(evaluation_function: Callable, true_array, pred_array):
+        def _get_index_list(
+            evaluation_function: Callable, true_array, pred_array, scheme_seqeval=None
+        ):
+            kwargs = (
+                {"mode": "strict", "scheme": scheme_seqeval}
+                if scheme_seqeval is not None
+                else {}
+            )
             try:
                 metric_list = evaluation_function(
                     true_array,
                     pred_array,
                     average=None,
                     zero_division="warn",
+                    **kwargs,
                 )
                 index_list = [i for i in range(len(metric_list))]
             except UndefinedMetricWarning:
                 metric_list_all = evaluation_function(
-                    true_array, pred_array, average=None, zero_division=0
+                    true_array,
+                    pred_array,
+                    average=None,
+                    zero_division=0,
+                    **kwargs,
                 )
 
                 index_list = list()
@@ -290,7 +395,11 @@ class NerMetrics:
                         index_list.append(index)
                     else:
                         metric_elem_alt = evaluation_function(
-                            true_array, pred_array, average=None, zero_division=1
+                            true_array,
+                            pred_array,
+                            average=None,
+                            zero_division=1,
+                            **kwargs,
                         )[index]
                         if metric_elem_alt != 1:
                             index_list.append(index)
@@ -313,11 +422,13 @@ class NerMetrics:
                 evaluation_function=precision_seqeval,
                 true_array=[self.true_flat_bio],
                 pred_array=[self.pred_flat_bio],
+                scheme_seqeval=self.scheme_entity_seqeval,
             )
             index_list_recall = _get_index_list(
                 evaluation_function=recall_seqeval,
                 true_array=[self.true_flat_bio],
                 pred_array=[self.pred_flat_bio],
+                scheme_seqeval=self.scheme_entity_seqeval,
             )
 
         self.results.classindices_macro = tuple(
@@ -331,15 +442,14 @@ class NerMetrics:
             self.results.numberofclasses_macro = len(self.results.classindices_macro)
 
     def _entity_evaluation_macro(
-        self, evaluation_function: Callable, restrict_macro: bool = True
+        self,
+        evaluation_function: Callable,
     ) -> float:
         """
         compute precision/recall macro average on entity level
 
         Args:
             evaluation_function: precision_seqeval, recall_seqeval
-            restrict_macro: if True,  compute macro-average only on well-defined labels
-                            if False, compute macro-average on all labels (replacing the ill-defined ones by 0)
 
         Returns:
             metric: precision/recall on entity level, 'macro' averaged on well-defined classes
@@ -349,43 +459,24 @@ class NerMetrics:
             recall_seqeval,
         ], f"evaluation function = {evaluation_function} unknown / not allowed."
 
-        if restrict_macro:
-            metric_list = evaluation_function(
-                [self.true_flat_bio],
-                [self.pred_flat_bio],
-                average=None,
-                zero_division=0,
-            )
-            metric_list_filtered = [
-                metric_list[index] for index in self.results.classindices_macro
-            ]
-            metric = (
-                np.average(metric_list_filtered)
-                if len(metric_list_filtered)
-                else self.failure_value
-            )
-        else:
-            try:
-                metric = evaluation_function(
-                    [self.true_flat_bio], [self.pred_flat_bio], average="macro"
-                )
-            except UndefinedMetricWarning as e:
-                if self.verbose:
-                    print(e)
-                metric = self.failure_value
-
+        metric = evaluation_function(
+            [self.true_flat_bio],
+            [self.pred_flat_bio],
+            mode="strict",
+            scheme=self.scheme_entity_seqeval,
+            average="macro",
+            zero_division=0,
+        )
         return metric
 
     def _entity_evaluation_f1(
-        self, evaluation_function: Callable, restrict_macro: bool = True
+        self, evaluation_function: Callable
     ) -> Tuple[float, float]:
         """
         compute f1 micro or macro average on entity level
 
         Args:
             evaluation_function: f1_seqeval
-            restrict_macro: if True,  compute macro-average only on well-defined labels
-                            if False, compute macro-average on all labels (replacing the ill-defined ones by 0)
 
         Returns:
             f1_micro: f1 on entity level, 'micro' averaged
@@ -406,17 +497,23 @@ class NerMetrics:
         ):
             f1_micro = self.failure_value
         else:
-            if self.tag_index is None:  # "fil"
+            if self.class_index is None:  # "fil"
                 f1_micro = evaluation_function(
-                    [self.true_flat_bio], [self.pred_flat_bio], average="micro"
+                    [self.true_flat_bio],
+                    [self.pred_flat_bio],
+                    average="micro",
+                    mode="strict",
+                    scheme=self.scheme_entity_seqeval,
                 )
             else:  # "ind"
                 f1_micro = evaluation_function(
                     [self.true_flat_bio],
                     [self.pred_flat_bio],
+                    mode="strict",
+                    scheme=self.scheme_entity_seqeval,
                     average=None,
                     zero_division="warn",
-                )[self.tag_index]
+                )[self.class_index]
 
         # f1_macro
         if (
@@ -425,21 +522,15 @@ class NerMetrics:
         ):
             f1_macro = self.failure_value
         else:
-            if self.tag_index is None:  # "fil"
-                if restrict_macro:
-                    metric_list = evaluation_function(
-                        [self.true_flat_bio],
-                        [self.pred_flat_bio],
-                        average=None,
-                    )
-                    metric_list_filtered = [
-                        metric_list[index] for index in self.results.classindices_macro
-                    ]
-                    f1_macro = np.average(metric_list_filtered)
-                else:
-                    f1_macro = evaluation_function(
-                        [self.true_flat_bio], [self.pred_flat_bio], average="macro"
-                    )
+            if self.class_index is None:  # "fil"
+                metric_list = evaluation_function(
+                    [self.true_flat_bio],
+                    [self.pred_flat_bio],
+                    mode="strict",
+                    scheme=self.scheme_entity_seqeval,
+                    average=None,
+                )
+                f1_macro = np.average(metric_list)
             else:  # "ind"
                 f1_macro = self.failure_value
 
@@ -457,107 +548,7 @@ class Results:
     f1_macro: float = -1
     classindices_macro: Tuple[float, ...] = ()
     numberofclasses_macro: float = -1
-
-
-def _assert_plain_tags(tag_list: List[str]) -> None:
-    for tag in tag_list:
-        if tag != "O" and (len(tag) > 2 and tag[1] == "-"):
-            raise Exception(
-                "ERROR! attempt to convert tags to bio format that already seem to have bio format."
-            )
-
-
-def _assert_bio_tags(tag_list: List[str]) -> None:
-    for tag in tag_list:
-        if tag != "O" and (len(tag) <= 2 or tag[1] != "-"):
-            raise Exception(
-                "ERROR! assuming tags to have bio format that seem to have plain format instead."
-            )
-
-
-def convert2bio(tag_list: List[str], convert_to_bio=True) -> List[str]:
-    """
-    - add bio prefixes if tag_list is in plain annotation scheme
-
-    Args:
-        tag_list:       e.g. ['O',   'ORG',   'ORG']
-        convert_to_bio: whether to cast to bio labels
-
-    Returns:
-        bio_tag_list:  e.g. ['O', 'B-ORG', 'I-ORG']
-    """
-    if convert_to_bio:
-        _assert_plain_tags(tag_list)
-        return _convert_tags_plain2bio(tag_list)
-    else:
-        _assert_bio_tags(tag_list)
-        return list(tag_list)
-
-
-def _convert_tags_plain2bio(tag_list: List[str]) -> List[str]:
-    """
-    adds bio prefixes to plain tags
-
-    Args:
-        tag_list:     e.g. ['O',   'ORG',   'ORG']
-
-    Returns:
-        bio_tag_list: e.g. ['O', 'B-ORG', 'I-ORG']
-    """
-    return [
-        _convert_tag_plain2bio(tag_list[i], previous=tag_list[i - 1] if i > 0 else None)
-        for i in range(len(tag_list))
-    ]
-
-
-def _convert_tag_plain2bio(tag: str, previous: Optional[str] = None) -> str:
-    """
-    add bio prefix to plain tag, depending on previous tag
-
-    Args:
-        tag:      e.g. 'ORG'
-        previous: e.g. 'ORG'
-
-    Returns:
-        bio_tag:  e.g. 'I-ORG'
-    """
-    if tag == "O" or tag.startswith("["):
-        return tag
-    elif previous is None:
-        return f"B-{tag}"
-    elif tag != previous:
-        return f"B-{tag}"
-    else:
-        return f"I-{tag}"
-
-
-def convert2plain(tag_list: List[str], convert_to_plain=True) -> List[str]:
-    """
-    - removes bio prefixes if tag_list is in bio annotation scheme
-
-    Args:
-        tag_list:  e.g. ['O', 'B-ORG', 'I-ORG']
-        convert_to_plain: whether to cast to plain labels
-
-    Returns:
-        tag_list_plain:       e.g. ['O',   'ORG',   'ORG']
-    """
-    if convert_to_plain:
-        _assert_bio_tags(tag_list)
-        return _convert_tags_bio2plain(tag_list)
-    else:
-        _assert_plain_tags(tag_list)
-        return list(tag_list)
-
-
-def _convert_tags_bio2plain(bio_tag_list: List[str]) -> List[str]:
-    """
-    retrieve plain tags by removing bio prefixes
-
-    Args:
-        bio_tag_list: e.g. ['O', 'B-ORG', 'I-ORG']
-
-    Returns:
-        tag_list:     e.g. ['O',   'ORG',   'ORG']
-    """
-    return [elem.split("-")[-1] for elem in bio_tag_list]
+    asr_abidance: float = -1
+    asr_precision_micro: float = -1
+    asr_recall_micro: float = -1
+    asr_f1_micro: float = -1
