@@ -5,8 +5,8 @@ import pytorch_lightning as pl
 from abc import ABC, abstractmethod
 import torch
 from torch.optim.optimizer import Optimizer
-from typing import List, Dict, Optional, Callable, Union, Any
-from omegaconf import OmegaConf
+from typing import List, Dict, Optional, Callable, Tuple, Any
+from omegaconf import DictConfig
 
 from torch.optim.lr_scheduler import LambdaLR
 
@@ -16,20 +16,20 @@ from transformers import get_constant_schedule_with_warmup
 from transformers import get_cosine_schedule_with_warmup
 from transformers import get_cosine_with_hard_restarts_schedule_with_warmup
 from transformers import AutoTokenizer
+from transformers import PreTrainedModel
 
 from nerblackbox.modules.ner_training.data_preprocessing.data_preprocessor import (
     DataPreprocessor,
 )
 from nerblackbox.modules.utils.util_functions import split_parameters
 from nerblackbox.modules.ner_training.ner_model_evaluation import NerModelEvaluation
+from nerblackbox.modules.ner_training.annotation_tags.annotation import Annotation
 from nerblackbox.modules.ner_training.logging.mlflow_client import MLflowClient
 from nerblackbox.modules.ner_training.logging.default_logger import DefaultLogger
 
-NEWLINE_TOKENS = ["[newline]", "[NEWLINE]"]
-
 
 class NerModel(pl.LightningModule, ABC):
-    def __init__(self, hparams: OmegaConf):
+    def __init__(self, hparams: DictConfig):
         """
         :param hparams: attr: experiment_name, run_name, pretrained_model_name, dataset_name, ..
         """
@@ -56,8 +56,9 @@ class NerModel(pl.LightningModule, ABC):
         :created attr: default_logger    [DefaultLogger]
         :created attr: logged_metrics    [LoggedMetrics]
         :created attr: tokenizer         [transformers AutoTokenizer]
+        :created attr: special_tokens    [list] of [str] e.g. ["[NEWLINE]", "[TAB]"]
         :created attr: data_preprocessor [DataPreprocessor]
-        :created attr: tag_list          [list] of tags in dataset, e.g. ['O', 'PER', 'LOC', ..]
+        :created attr: annotation        [Annotation]
         :created attr: model             [transformers AutoModelForTokenClassification]
 
         [only train]
@@ -70,10 +71,12 @@ class NerModel(pl.LightningModule, ABC):
         :created attr: scheduler              [torch LambdaLR]
         :return: -
         """
-        self.mlflow_client: MLflowClient
         self.default_logger: DefaultLogger
+        self.annotation: Annotation
+        self.model: PreTrainedModel
+        self.mlflow_client: MLflowClient
+        self.optimizer: Optimizer
         self.scheduler: LambdaLR
-        self.tag_list: List[str]
 
     def _preparations_data_general(self):
         """
@@ -101,7 +104,7 @@ class NerModel(pl.LightningModule, ABC):
         self.tokenizer = AutoTokenizer.from_pretrained(
             self.pretrained_model_name,
             do_lower_case=False,
-            additional_special_tokens=NEWLINE_TOKENS,
+            additional_special_tokens=self.special_tokens,
             use_fast=True,
         )  # do_lower_case needs to be False !!
 
@@ -119,35 +122,35 @@ class NerModel(pl.LightningModule, ABC):
     ####################################################################################################################
     # FORWARD & BACKWARD PROPAGATION
     ####################################################################################################################
-    def forward(self, _input_ids, _attention_mask, _segment_ids, _tag_ids):
+    def forward(self, *args, **kwargs) -> List:
         """
-        :param _input_ids:            [torch tensor] of shape [batch_size, seq_length],
-                                             e.g. 1st row = [1, 567, 568, 569, .., 2, 611, 612, .., 2, 0, 0, 0, ..]
-        :param _attention_mask:       [torch tensor] of shape [batch_size, seq_length],
-                                             e.g. 1st row = [1,   1,   1,   1, .., 1,   1,   1, .., 1, 0, 0, 0, ..]
-        :param _segment_ids:          [torch tensor] of shape [batch_size, seq_length],
-                                             e.g. 1st row = [0,   0,   0,   0, .., 0,   1,   1, .., 1, 0, 0, 0, ..]
-        :param _tag_ids:              [torch tensor] of shape [batch_size, seq_length],
-                                             e.g. 1st row = [1,   3,   3,   4, .., 2,   3,   3, .., 2, 0, 0, 0, ..]
-        :return: _outputs: [list] of 2 elements:
-                    i)  _loss: [float] cross entropy between _tag_ids & _tags_ids_predictions
-                                       on non-padding tokens (i.e. where elements in _input_ids are not 0)
-                    ii) _tag_ids_prediction_logits: [torch tensor] of shape [batch_size, seq_length, vocabulary_size]
+        Args:
+            args = (batch),
+            where batch = Dict[str, torch.Tensor]
+            with keys = subset of EncodingsKeys,
+                 values = 2D torch tensor of shape [batch_size, seq_length]
+            e.g.
+            input_ids      = [2D torch tensor], e.g. [[1, 567, 568, 569, .., 2, 611, 612, .., 2, 0, 0, 0, ..], [..], ..]
+            attention_mask = [2D torch tensor], e.g. [[1,   1,   1,   1, .., 1,   1,   1, .., 1, 0, 0, 0, ..], [..], ..]
+            token_type_ids = [2D torch tensor], e.g. [[0,   0,   0,   0, .., 0,   1,   1, .., 1, 0, 0, 0, ..], [..], ..]
+            labels         = [2D torch tensor], e.g. [[1,   3,   3,   4, .., 2,   3,   3, .., 2, 0, 0, 0, ..], [..], ..]
+
+        Returns:
+            _outputs: [list] of 1 or 2 elements:
+                   i)  _loss: [float] cross entropy between _labels & _tags_ids_predictions
+                                      on non-padding tokens (i.e. where elements in _input_ids are not 0)
+                   ii) _labels_prediction_logits: [torch tensor] of shape [batch_size, seq_length, vocabulary_size],
+                       if labels are provided in _batch
         """
-        return self.model(
-            input_ids=_input_ids,
-            attention_mask=_attention_mask,
-            token_type_ids=_segment_ids,
-            labels=_tag_ids,
-        )
+        return self.model(**args[0])
 
     ####################################################################################################################
     # TRAIN
     ####################################################################################################################
     def training_step(self, batch, batch_idx):
         # REQUIRED
-        input_ids, attention_mask, segment_ids, tag_ids = batch
-        outputs = self(input_ids, attention_mask, segment_ids, tag_ids)
+        input_ids, attention_mask, token_type_ids, labels = self._parse_batch(batch)
+        outputs = self(batch)
         batch_train_loss = outputs[0]
 
         # logging
@@ -156,7 +159,13 @@ class NerModel(pl.LightningModule, ABC):
         # debug
         if batch_idx == 0:
             self._debug_step_check(
-                "train", batch, outputs, input_ids, attention_mask, segment_ids, tag_ids
+                "train",
+                batch,
+                outputs,
+                input_ids,
+                attention_mask,
+                token_type_ids,
+                labels,
             )
 
         return {"loss": batch_train_loss}
@@ -194,50 +203,65 @@ class NerModel(pl.LightningModule, ABC):
     ####################################################################################################################
     def validation_step(self, batch, batch_idx):
         # OPTIONAL
-        input_ids, attention_mask, segment_ids, tag_ids = batch
-        outputs = self(input_ids, attention_mask, segment_ids, tag_ids)
+        input_ids, attention_mask, token_type_ids, labels = self._parse_batch(batch)
+        outputs = self(batch)
         batch_loss, logits = outputs[:2]
 
         # debug
         if batch_idx == 0:
             self._debug_step_check(
-                "val", batch, outputs, input_ids, attention_mask, segment_ids, tag_ids
+                "val", batch, outputs, input_ids, attention_mask, token_type_ids, labels
             )
 
-        return batch_loss, tag_ids, logits
+        return batch_loss, labels, logits
 
     def validation_epoch_end(self, outputs):
         """
-        :param outputs: [list] of [list] w/ 3 elements [batch_loss, batch_tag_ids, batch_logits] for each batch
+        :param outputs: [list] of [list] w/ 3 elements [batch_loss, batch_labels, batch_logits] for each batch
         :return:        [dict] w/ key 'val_loss' & value = mean batch loss of val dataset [float]
         """
         # OPTIONAL
-        return self._validate_on_epoch("val", outputs=outputs)
+        if self.trainer.state.fn == "fit":
+            metrics = "loss"
+        elif self.trainer.state.fn == "validate":
+            metrics = "all"
+        else:
+            raise Exception(
+                f"ERROR! self.trainer.state.fn = {self.trainer.state.fn} unexpected."
+            )
+
+        return self._validate_on_epoch("val", outputs=outputs, metrics=metrics)
 
     ####################################################################################################################
     # TEST
     ####################################################################################################################
     def test_step(self, batch, batch_idx):
         # OPTIONAL
-        input_ids, attention_mask, segment_ids, tag_ids = batch
-        outputs = self(input_ids, attention_mask, segment_ids, tag_ids)
+        input_ids, attention_mask, token_type_ids, labels = self._parse_batch(batch)
+        outputs = self(batch)
         batch_loss, logits = outputs[:2]
 
         # debug
         if batch_idx == 0:
             self._debug_step_check(
-                "test", batch, outputs, input_ids, attention_mask, segment_ids, tag_ids
+                "test",
+                batch,
+                outputs,
+                input_ids,
+                attention_mask,
+                token_type_ids,
+                labels,
             )
 
-        return batch_loss, tag_ids, logits
+        return batch_loss, labels, logits
 
     def test_epoch_end(self, outputs):
         """
-        :param outputs: [list] of [list] w/ 3 elements [batch_loss, batch_tag_ids, batch_logits] for each batch
+        :param outputs: [list] of [list] w/ 3 elements [batch_loss, batch_labels, batch_logits] for each batch
         :return:        [dict] w/ key 'test_loss' & value = mean batch loss of test dataset [float]
         """
         # OPTIONAL
-        return self._validate_on_epoch("test", outputs=outputs)
+        return self._validate_on_epoch("test", outputs=outputs, metrics="all")
 
     ####################################################################################################################
     # DATALOADER
@@ -265,25 +289,30 @@ class NerModel(pl.LightningModule, ABC):
     # 1. PREPARATIONS
     ####################################################################################################################
     def _create_optimizer(
-        self, learning_rate, fp16=True, no_decay=("bias", "gamma", "beta")
-    ):
+        self,
+        learning_rate: float,
+        fp16: bool = True,
+        no_decay: Tuple[str, ...] = ("bias", "LayerNorm.weight"),
+    ) -> Optimizer:
         """
         create optimizer with basic learning rate and L2 normalization for some parameters
-        ----------------------------------------------------------------------------------
-        :param learning_rate: [float] basic learning rate
-        :param fp16:          [bool]
-        :param no_decay:      [tuple of str] parameters that contain one of those are not subject to L2 normalization
-        :return: optimizer:   [torch optimizer]
+
+        Args:
+            learning_rate: [float] basic learning rate
+            fp16:          [bool]
+            no_decay:      [tuple of str] parameters that contain one of those are not subject to L2 normalization
+
+        Returns:
+            optimizer:   [torch optimizer]
         """
         # Remove unused pooler that otherwise break Apex
         param_optimizer = list(self.model.named_parameters())
-        param_optimizer = [n for n in param_optimizer if "pooler" not in n[0]]
-        optimizer_grouped_parameters = [
+        optimizer_grouped_parameters: List[Any] = [
             {
                 "params": [
                     p for n, p in param_optimizer if not any(nd in n for nd in no_decay)
                 ],
-                "weight_decay": 0.02,
+                "weight_decay": 0.01,
             },
             {
                 "params": [
@@ -292,15 +321,21 @@ class NerModel(pl.LightningModule, ABC):
                 "weight_decay": 0.0,
             },
         ]
-        # print('> param_optimizer')
-        # print([n for n, p in param_optimizer])
+        # self.default_logger.log_debug(
+        #     '> param_optimizer:',
+        #     [n for n, p in param_optimizer],
+        # )
         self.default_logger.log_debug(
             "> parameters w/  weight decay:",
-            len(optimizer_grouped_parameters[0]["params"]),
+            len(optimizer_grouped_parameters[0]["params"])
+            if isinstance(optimizer_grouped_parameters[0]["params"], list)
+            else None,
         )
         self.default_logger.log_debug(
             "> parameters w/o weight decay:",
-            len(optimizer_grouped_parameters[1]["params"]),
+            len(optimizer_grouped_parameters[1]["params"])
+            if isinstance(optimizer_grouped_parameters[1]["params"], list)
+            else None,
         )
         if fp16:
             optimizer = AdamW(optimizer_grouped_parameters, lr=learning_rate)
@@ -315,13 +350,14 @@ class NerModel(pl.LightningModule, ABC):
         return optimizer
 
     def _create_scheduler(
-        self, _lr_warmup_epochs, _lr_schedule, _lr_num_cycles=None
+        self, _lr_warmup_epochs, _lr_schedule, _lr_max_epochs=None, _lr_num_cycles=None
     ) -> LambdaLR:
         """
         create scheduler with warmup
         ----------------------------
         :param _lr_warmup_epochs:   [int]
-        :param _lr_schedule:        [str], 'linear', 'constant', 'cosine', 'cosine_with_hard_resets'
+        :param _lr_schedule:        [str], 'linear', 'constant', 'cosine', 'cosine_with_hard_resets', 'hybrid'
+        :param _lr_max_epochs:      [int, optional] needed for _lr_schedule != 'constant'
         :param _lr_num_cycles:      [float, optional], e.g. 0.5, 1.0, only for cosine learning rate schedules
         :return: scheduler          [torch LambdaLR] learning rate scheduler
         """
@@ -330,10 +366,10 @@ class NerModel(pl.LightningModule, ABC):
             "linear",
             "cosine",
             "cosine_with_hard_restarts",
+            "hybrid",
         ]:
             raise Exception(f"lr_schedule = {_lr_schedule} not implemented.")
 
-        num_training_steps = self._get_steps(self.hyperparameters.max_epochs)
         num_warmup_steps = self._get_steps(_lr_warmup_epochs)
 
         scheduler_params = {
@@ -341,9 +377,13 @@ class NerModel(pl.LightningModule, ABC):
             "last_epoch": -1,
         }
 
-        if _lr_schedule == "constant":
+        if _lr_schedule in ["constant", "hybrid"]:
             return get_constant_schedule_with_warmup(self.optimizer, **scheduler_params)
         else:
+            assert (
+                _lr_max_epochs is not None
+            ), f"ERROR! need to specify _lr_max_epochs for _lr_schedule = {_lr_schedule}"
+            num_training_steps = self._get_steps(_lr_max_epochs)
             scheduler_params["num_training_steps"] = num_training_steps
 
             if _lr_schedule == "linear":
@@ -381,39 +421,80 @@ class NerModel(pl.LightningModule, ABC):
         """
         return _num_epochs * len(self.dataloader["train"])
 
+    @staticmethod
+    def _parse_batch(_batch) -> Tuple[torch.Tensor, ...]:
+        """
+        Args:
+            _batch: Dict with keys = subset of EncodingsKeys, values = 2D torch tensor of shape [batch_size, seq_length]
+            e.g.
+            input_ids      = [2D torch tensor], e.g. [[1, 567, 568, 569, .., 2, 611, 612, .., 2, 0, 0, 0, ..], [..], ..]
+            attention_mask = [2D torch tensor], e.g. [[1,   1,   1,   1, .., 1,   1,   1, .., 1, 0, 0, 0, ..], [..], ..]
+            token_type_ids = [2D torch tensor], e.g. [[0,   0,   0,   0, .., 0,   1,   1, .., 1, 0, 0, 0, ..], [..], ..]
+            labels         = [2D torch tensor], e.g. [[1,   3,   3,   4, .., 2,   3,   3, .., 2, 0, 0, 0, ..], [..], ..]
+
+        Returns:
+            input_ids      = [2D torch tensor], e.g. [[1, 567, 568, 569, .., 2, 611, 612, .., 2, 0, 0, 0, ..], [..], ..]
+            attention_mask = [2D torch tensor], e.g. [[1,   1,   1,   1, .., 1,   1,   1, .., 1, 0, 0, 0, ..], [..], ..]
+            token_type_ids = [2D torch tensor], e.g. [[0,   0,   0,   0, .., 0,   1,   1, .., 1, 0, 0, 0, ..], [..], ..]
+            labels         = [2D torch tensor], e.g. [[1,   3,   3,   4, .., 2,   3,   3, .., 2, 0, 0, 0, ..], [..], ..]
+        """
+        input_ids = _batch["input_ids"]
+        attention_mask = _batch["attention_mask"]
+        token_type_ids = (
+            _batch["token_type_ids"] if "token_type_ids" in _batch.keys() else None
+        )
+        labels = _batch["labels"]
+        return input_ids, attention_mask, token_type_ids, labels
+
+    ####################################################################################################################
+    # 2. VALIDATE / COMPUTE METRICS
+    ####################################################################################################################
     def _validate_on_epoch(
-        self, phase: str, outputs: List[Union[torch.Tensor, Dict[str, Any]]]
+        self,
+        phase: str,
+        outputs: List[torch.Tensor],
+        metrics: str,
     ) -> Dict[str, float]:
         """
         Args:
             phase:   [str], 'val', 'test'
-            outputs: [list] of [lists] = [batch_loss, batch_tag_ids, batch_logits] with 3 torch tensors for each batch
+            outputs: [list] of [lists] = [batch_loss, batch_labels, batch_logits] with 3 torch tensors for each batch
+            metrics: [str], 'loss', 'all'
 
         Returns:
             [dict] w/ key '<phase>_loss' & value = mean batch loss [float]
         """
-        ner_model_evaluation = NerModelEvaluation(
-            current_epoch=self.current_epoch,
-            tag_list=self.tag_list,
-            annotation_scheme=self.params.dataset_tags
-            if self.params.dataset_tags != "iob"
-            else "bio",  # TODO: clean
-            default_logger=self.default_logger,
-            logged_metrics=self.logged_metrics,
-        )
-        (
-            epoch_metrics,
-            classification_report,
-            confusion_matrix,
-            epoch_loss,
-        ) = ner_model_evaluation.execute(phase, outputs)
-        self._log_metrics_confusion_matrix_classification_report(
-            phase,
-            epoch_metrics,
-            confusion_matrix,
-            classification_report,
-        )
-        self.log(f"{phase}_loss", epoch_loss)  # for early stopping callback
+        if metrics == "loss":
+            # CPU (numpy)
+            # batch_loss = [output[0].detach().cpu().numpy() for output in outputs]
+            # epoch_loss = np.stack(batch_loss).mean()
+
+            # GPU (pytorch)
+            batch_loss = [output[0].detach() for output in outputs]
+            epoch_loss = torch.mean(torch.stack(batch_loss)).item()  # -> float
+
+            self.log(f"{phase}_loss", epoch_loss)  # for early stopping callback
+        elif metrics == "all":
+            ner_model_evaluation = NerModelEvaluation(
+                current_epoch=self.current_epoch,
+                annotation=self.annotation,
+                default_logger=self.default_logger,
+                logged_metrics=self.logged_metrics,
+            )
+            (
+                epoch_metrics,
+                classification_report,
+                confusion_matrix,
+                epoch_loss,
+            ) = ner_model_evaluation.execute(phase, outputs)
+            self._log_metrics_confusion_matrix_classification_report(
+                phase,
+                epoch_metrics,
+                confusion_matrix,
+                classification_report,
+            )
+        else:
+            raise Exception(f"ERROR! metrics = {metrics} unexpected.")
 
         return {f"{phase}_loss": epoch_loss}
 
@@ -504,24 +585,25 @@ class NerModel(pl.LightningModule, ABC):
         _outputs,
         _input_ids,
         _attention_mask,
-        _segment_ids,
-        _tag_ids,
+        _token_type_ids,
+        _labels,
     ):
         self.default_logger.log_debug(f"{phase.upper()} STEP CHECK")
-        self.default_logger.log_debug(f"batch on gpu:   {_batch[0].is_cuda}")
+        self.default_logger.log_debug(f"batch on gpu:   {_batch['input_ids'].is_cuda}")
         self.default_logger.log_debug(f"outputs on gpu: {_outputs[0].is_cuda}")
-        self.default_logger.log_debug(
-            f"input_ids      shape|1st row: {_input_ids.shape} | {_input_ids[0]}"
-        )
-        self.default_logger.log_debug(
-            f"attention_mask shape|1st row: {_attention_mask.shape} | {_attention_mask[0]}"
-        )
-        self.default_logger.log_debug(
-            f"segment_ids    shape|1st row: {_segment_ids.shape} | {_segment_ids[0]}"
-        )
-        self.default_logger.log_debug(
-            f"tag_ids        shape|1st row: {_tag_ids.shape} | {_tag_ids[0]}"
-        )
+        name_tensor_combinations = [
+            ("input_ids", _input_ids),
+            ("attention_mask", _attention_mask),
+            ("token_type_ids", _token_type_ids),
+            ("labels", _labels),
+        ]
+        for _name, _tensor in name_tensor_combinations:
+            if _tensor is not None:
+                self.default_logger.log_debug(
+                    f"{_name.ljust(15)} shape|1st row: {_tensor.shape} | {_tensor[0]}"
+                )
+            else:
+                self.default_logger.log_debug(f"{_name.ljust(15)} = None")
 
     def _print_metrics(self, phase, _metrics, _classification_reports: Optional[str]):
         """
