@@ -1,12 +1,16 @@
 import json
 
+import torch
 import numpy as np
 from transformers import AutoModelForTokenClassification
 from torch.nn.functional import softmax
-from typing import List, Union, Tuple, Any, Dict, IO
-from omegaconf import OmegaConf
+from typing import List, Union, Tuple, Any, Dict, IO, Optional, Callable
+from omegaconf import DictConfig
 
 from nerblackbox.modules.ner_training.ner_model import NerModel
+from nerblackbox.modules.ner_training.annotation_tags.annotation import Annotation
+from nerblackbox.modules.ner_training.annotation_tags.token_tags import TokenTags
+from nerblackbox.tests.utils import PseudoDefaultLogger
 
 from transformers import logging
 
@@ -15,6 +19,8 @@ logging.set_verbosity_error()
 
 VERBOSE = False
 
+PREDICTIONS = List[List[Dict[str, Union[str, Dict]]]]
+
 
 class NerModelPredict(NerModel):
     """
@@ -22,20 +28,32 @@ class NerModelPredict(NerModel):
     """
 
     @classmethod
-    def load_from_checkpoint(cls, checkpoint_path: Union[str, IO]) -> "NerModelPredict":
+    def load_from_checkpoint(
+        cls,
+        checkpoint_path: Union[str, IO],
+        map_location: Optional[
+            Union[Dict[str, str], str, torch.device, int, Callable]
+        ] = None,
+        hparams_file: Optional[str] = None,
+        strict: bool = True,
+        **kwargs,
+    ) -> "NerModelPredict":
         """load model in inference mode from checkpoint_path
 
         Args:
             checkpoint_path: path to checkpoint
+            map_location: see ModelIO
+            hparams_file: see ModelIO
+            strict: see ModelIO
 
         Returns:
             model loaded from checkpoint
         """
-        model = super().load_from_checkpoint(checkpoint_path)
+        model = super().load_from_checkpoint(checkpoint_path, **kwargs)
         model.freeze()  # for inference mode
         return model
 
-    def __init__(self, hparams: OmegaConf):
+    def __init__(self, hparams: DictConfig):
         """
         Args:
             hparams: attr experiment_name, run_name, pretrained_model_name, dataset_name, ..
@@ -50,41 +68,46 @@ class NerModelPredict(NerModel):
         :created attr: default_logger    [DefaultLogger]
         :created attr: logged_metrics    [LoggedMetrics]
         :created attr: tokenizer         [transformers AutoTokenizer]
+        :created attr: special_tokens    [list] of [str] e.g. ["[NEWLINE]", "[TAB]"]
         :created attr: data_preprocessor [DataPreprocessor]
-        :created attr: tag_list          [list] of tags in dataset, e.g. ['O', 'PER', 'LOC', ..]
+        :created attr: annotation        [Annotation]
         :created attr: model             [transformers AutoModelForTokenClassification]
         :return: -
         """
         # predict
         self._preparations_predict()  # attr: default_logger
         self._preparations_data_general()  # attr: tokenizer, data_preprocessor
-        self._preparations_data_predict()  # attr: tag_list, model
+        self._preparations_data_predict()  # attr: annotation, model
 
     def _preparations_predict(self):
         """
-        :created attr: default_logger         [None]
+        :created attr: default_logger    [None]
+        :created attr: special_tokens    [list] of [str] e.g. ["[NEWLINE]", "[TAB]"]
         :return: -
         """
-        self.default_logger = None
+        self.default_logger = PseudoDefaultLogger()
+
+        # special tokens
+        self.special_tokens = json.loads(self.hparams.special_tokens)
 
     def _preparations_data_predict(self):
         """
-        :created attr: tag_list          [list] of tags in dataset, e.g. ['O', 'PER', 'LOC', ..]
+        :created attr: annotation        [Annotation]
         :created attr: model             [transformers AutoModelForTokenClassification]
         :return: -
         """
-        # tag_list
-        self.tag_list = json.loads(self.hparams.tag_list)
+        # annotation
+        self.annotation = Annotation(json.loads(self.hparams.annotation_classes))
 
         # model
         self.model = AutoModelForTokenClassification.from_pretrained(
             self.pretrained_model_name,
-            num_labels=len(self.tag_list),
+            num_labels=len(self.annotation.classes),
             return_dict=False,
         )
         self.model.resize_token_embeddings(
             len(self.tokenizer)
-        )  # due to addtional_special_tokens = NEWLINE_TOKENS
+        )  # due to addtional_special_tokens
 
     ####################################################################################################################
     # PREDICT
@@ -93,41 +116,47 @@ class NerModelPredict(NerModel):
         self,
         input_texts: Union[str, List[str]],
         level: str = "entity",
-        autocorrect: bool = True,
-    ) -> List[List[Dict[str, str]]]:
+        autocorrect: bool = False,
+    ) -> PREDICTIONS:
         """predict tags for input texts. output on entity or word level.
 
         Examples:
-            ```
-            predict(["arbetsförmedlingen finns i stockholm"], level="entity")
-            # [
-            #     {"char_start": "0", "char_end": "18", "token": "arbetsförmedlingen", "tag": "ORG"},
-            #     {"char_start": "27", "char_end": "36", "token": "stockholm", "tag": "LOC"},
-            # ]
-            ```
             ```
             predict(["arbetsförmedlingen finns i stockholm"], level="word", autocorrect=False)
-            # [
+            # [[
             #     {"char_start": "0", "char_end": "18", "token": "arbetsförmedlingen", "tag": "I-ORG"},
             #     {"char_start": "19", "char_end": "24", "token": "finns", "tag": "O"},
             #     {"char_start": "25", "char_end": "26", "token": "i", "tag": "O"},
             #     {"char_start": "27", "char_end": "36", "token": "stockholm", "tag": "B-LOC"},
-            # ]
+            # ]]
             ```
             ```
             predict(["arbetsförmedlingen finns i stockholm"], level="word", autocorrect=True)
-            # [
+            # [[
             #     {"char_start": "0", "char_end": "18", "token": "arbetsförmedlingen", "tag": "B-ORG"},
             #     {"char_start": "19", "char_end": "24", "token": "finns", "tag": "O"},
             #     {"char_start": "25", "char_end": "26", "token": "i", "tag": "O"},
             #     {"char_start": "27", "char_end": "36", "token": "stockholm", "tag": "B-LOC"},
-            # ]
+            # ]]
+            ```
+            ```
+            predict(["arbetsförmedlingen finns i stockholm"], level="entity", autocorrect=False)
+            # [[
+            #     {"char_start": "27", "char_end": "36", "token": "stockholm", "tag": "LOC"},
+            # ]]
+            ```
+            ```
+            predict(["arbetsförmedlingen finns i stockholm"], level="entity", autocorrect=True)
+            # [[
+            #     {"char_start": "0", "char_end": "18", "token": "arbetsförmedlingen", "tag": "ORG"},
+            #     {"char_start": "27", "char_end": "36", "token": "stockholm", "tag": "LOC"},
+            # ]]
             ```
 
         Args:
             input_texts:   e.g. ["example 1", "example 2"]
             level:         "entity" or "word"
-            autocorrect:   if True, autocorrect annotation scheme (e.g. B- and I- tags). needs to be True if level == "entity".
+            autocorrect:   if True, autocorrect annotation scheme (e.g. B- and I- tags).
 
         Returns:
             predictions: [list] of predictions for the different examples.
@@ -135,20 +164,18 @@ class NerModelPredict(NerModel):
         """
         return self._predict(input_texts, level, autocorrect, proba=False)
 
-    def predict_proba(
-        self, input_texts: Union[str, List[str]]
-    ) -> List[List[Dict[str, Union[str, Dict]]]]:
+    def predict_proba(self, input_texts: Union[str, List[str]]) -> PREDICTIONS:
         """predict probability distributions for input texts. output on word level.
 
         Examples:
             ```
             predict_proba(["arbetsförmedlingen finns i stockholm"])
-            # [
+            # [[
             #     {"char_start": "0", "char_end": "18", "token": "arbetsförmedlingen", "proba_dist: {"O": 0.21, "B-ORG": 0.56, ..}},
             #     {"char_start": "19", "char_end": "24", "token": "finns", "proba_dist: {"O": 0.87, "B-ORG": 0.02, ..}},
             #     {"char_start": "25", "char_end": "26", "token": "i", "proba_dist: {"O": 0.95, "B-ORG": 0.01, ..}},
             #     {"char_start": "27", "char_end": "36", "token": "stockholm", "proba_dist: {"O": 0.14, "B-ORG": 0.22, ..}},
-            # ]
+            # ]]
             ```
 
         Args:
@@ -157,7 +184,7 @@ class NerModelPredict(NerModel):
         Returns:
             predictions: [list] of probability predictions for different examples.
                          each list contains a [list] of [dict] w/ keys = char_start, char_end, word, proba_dist
-                         where proba_dist = [dict] that maps self.tag_list to probabilities
+                         where proba_dist = [dict] that maps self.annotation.classes to probabilities
         """
         return self._predict(input_texts, level="word", autocorrect=False, proba=True)
 
@@ -165,30 +192,26 @@ class NerModelPredict(NerModel):
         self,
         input_texts: Union[str, List[str]],
         level: str = "entity",
-        autocorrect: bool = True,
+        autocorrect: bool = False,
         proba: bool = False,
-    ) -> List[List[Dict[str, Union[str, Dict]]]]:
+    ) -> PREDICTIONS:
         """predict tags or probabilities for tags
 
         Args:
             input_texts:  e.g. ["example 1", "example 2"]
             level:        "entity" or "word"
-            autocorrect:  if True, autocorrect annotation scheme (e.g. B- and I- tags). needs to be True if level == "entity".
+            autocorrect:  if True, autocorrect annotation scheme (e.g. B- and I- tags).
             proba:        if True, predict probabilities instead of labels (on word level)
 
         Returns:
             predictions: [list] of [list] of [dict] w/ keys = char_start, char_end, word, tag/proba_dist
-                         where proba_dist = [dict] that maps self.tag_list to probabilities
+                         where proba_dist = [dict] that maps self.annotation.classes to probabilities
         """
         # --- check input arguments ---
         assert level in [
             "entity",
             "word",
         ], f"ERROR! model prediction level = {level} unknown, needs to be entity or word."
-        if level == "entity":
-            assert (
-                autocorrect is True
-            ), f"ERROR! level = entity requires autocorrect = True."
         if proba:
             assert level == "word" and autocorrect is False, (
                 f"ERROR! probability predictions require level = word and autocorrect = False. "
@@ -237,19 +260,39 @@ class NerModelPredict(NerModel):
                 Dict[str, Union[str, Dict]]
             ] = restore_unknown_tokens(_input_text_word_predictions, input_text)
 
-            if autocorrect:
-                input_text_word_predictions = restore_annotation_scheme_consistency(
+            if autocorrect or level == "entity":
+                assert (
+                    proba is False
+                ), f"ERROR! autocorrect = {autocorrect} / level = {level} not allowed if proba = {proba}"
+
+                def assert_typing(input_text_word_predictions_str_dict):
+                    # this is only to ensure correct typing, it does not actually change anything
+                    return [
+                        {
+                            k: str(v)
+                            for k, v in input_text_word_prediction_str_dict.items()
+                        }
+                        for input_text_word_prediction_str_dict in input_text_word_predictions_str_dict
+                    ]
+
+                input_text_word_predictions_str: List[Dict[str, str]] = assert_typing(
                     input_text_word_predictions
                 )
 
-            if level == "entity":
-                assert (
-                    proba is False
-                ), f"ERROR! level = entity not allowed if proba = {proba}"
-                input_text_word_predictions = merge_tokens_to_entities(
-                    input_text_word_predictions, input_text
+                token_tags = TokenTags(
+                    input_text_word_predictions_str, scheme=self.annotation.scheme
                 )
 
+                if autocorrect:
+                    token_tags.restore_annotation_scheme_consistency()
+
+                if level == "entity":
+                    token_tags.merge_tokens_to_entities(
+                        original_text=input_text, verbose=VERBOSE
+                    )
+
+                input_text_word_predictions = token_tags.as_list()
+
             predictions.append(input_text_word_predictions)
 
         ######################################
@@ -272,7 +315,7 @@ class NerModelPredict(NerModel):
 
         # dataloader
         dataloader = self.data_preprocessor.to_dataloader(
-            input_examples, self.tag_list, batch_size=1
+            input_examples, self.annotation.classes, batch_size=1
         )
 
         return dataloader["predict"]
@@ -283,17 +326,10 @@ class NerModelPredict(NerModel):
         :return: output_token_tensors [list] of [torch tensor] of shape [1, #tags]
         :return: tokens               [list] of [str]
         """
-        (
-            input_ids,  # shape: [1, seq_length]
-            attention_mask,  # shape: [1, seq_length]
-            segment_ids,  # shape: [1, seq_length]
-            label_ids,  # shape: [1, seq_length]
-        ) = sample
-
+        input_ids, attention_mask, token_type_ids, labels = self._parse_batch(sample)
+        sample.pop("labels")
         output = self.model(
-            input_ids=input_ids,
-            attention_mask=attention_mask,
-            token_type_ids=segment_ids,
+            **sample
         )  # shape: [1 (=#examples), 1 (=#batch_size), seq_length, #tags]
 
         output_token_tensors = [
@@ -311,7 +347,9 @@ class NerModelPredict(NerModel):
         :return: output_token_predictions [list] of [str]
         """
         return [
-            self.tag_list[np.argmax(output_token_tensors[i].detach().numpy())]
+            self.annotation.classes[
+                int(np.argmax(output_token_tensors[i].detach().numpy()))
+            ]
             for i in range(self._hparams.max_seq_length)
         ]
 
@@ -328,10 +366,10 @@ class NerModelPredict(NerModel):
 
         tag_probability_distribution = [
             {
-                self.tag_list[j]: float(
+                self.annotation.classes[j]: float(
                     probability_distributions[i][j].detach().numpy()
                 )
-                for j in range(len(self.tag_list))
+                for j in range(len(self.annotation.classes))
             }
             for i in range(self._hparams.max_seq_length)
         ]
@@ -479,167 +517,3 @@ def restore_unknown_tokens(
                 )
 
     return example_word_predictions_restored
-
-
-def restore_annotation_scheme_consistency(
-    example_word_predictions: List[Dict[str, Union[str, Dict]]],
-) -> List[Dict[str, Union[str, Dict]]]:
-    """
-    restore annotation scheme consistency in case of BIO tags
-    plain tags are not modified
-
-    Args:
-        example_word_predictions: e.g. [
-            {"char_start": "0", "char_end": "7", "token": "example", "tag": "I-TAG"},
-            ..
-        ]
-
-    Returns:
-        example_word_predictions_restored: e.g. [
-            {"char_start": "0", "char_end": "7", "token": "example", "tag": "B-TAG"},
-            ..
-        ]
-    """
-    example_word_predictions_restored: List[Dict[str, Union[str, Dict]]] = list()
-    for i in range(len(example_word_predictions)):
-        example_word_prediction_restored = example_word_predictions[i]
-        current_tag = example_word_predictions[i]["tag"]
-        current_tag = assert_str(current_tag, "current_tag")
-
-        if current_tag == "O" or "-" not in current_tag or current_tag.startswith("B-"):
-            example_word_predictions_restored.append(example_word_prediction_restored)
-        elif current_tag.startswith("I-"):
-            previous_tag = example_word_predictions[i - 1]["tag"] if i > 0 else None
-
-            if previous_tag not in [current_tag, current_tag.replace("I-", "B-")]:
-                example_word_prediction_restored["tag"] = current_tag.replace(
-                    "I-", "B-"
-                )
-
-            example_word_predictions_restored.append(example_word_prediction_restored)
-        else:
-            raise Exception(
-                f"ERROR! current tag = {current_tag} expected to be of the form I-*"
-            )
-
-    assert len(example_word_predictions_restored) == len(
-        example_word_predictions
-    ), f"ERROR!"
-
-    return example_word_predictions_restored
-
-
-def merge_tokens_to_entities(
-    example_word_predictions: List[Dict[str, Union[str, Dict]]],
-    example: str,
-) -> List[Dict[str, Union[str, Dict]]]:
-    """
-    merge token predictions that belong together (B-* & I-*)
-    discard tokens with tag 'O'
-
-    Args:
-        example_word_predictions: e.g. [
-            {"char_start": "0", "char_end": "7", "token": "example", "tag": "B-TAG"},
-            {"char_start": "8", "char_end": "16", "token": "sentence", "tag": "I-TAG"},
-            {"char_start": "17", "char_end": "18", "token": ".", "tag": "O"},
-            ..
-        ]
-        example: str
-
-    Returns:
-        example_word_predictions_merged: e.g. [
-            {"char_start": "0", "char_end": "16", "token": "example", "tag": "TAG"},
-            ..
-        ]
-    """
-    merged_ner_tags = list()
-    count = {
-        "o_tags": 0,
-        "replace": 0,
-        "merge": 0,
-        "unmodified": 0,
-    }
-    for i in range(len(example_word_predictions)):
-        current_tag = example_word_predictions[i]["tag"]
-        current_tag = assert_str(current_tag, "current_tag")
-        if current_tag == "O":
-            # merged_ner_tag = example_word_predictions[i]
-            count["o_tags"] += 1
-            # merged_ner_tags.append(merged_ner_tag)
-        elif current_tag.startswith("B-"):  # BIO scheme
-            n_tags = 0
-            for n in range(i + 1, len(example_word_predictions)):
-                next_tag = example_word_predictions[n]["tag"]
-                next_tag = assert_str(next_tag, "next_tag")
-                # next_token_start = example_word_predictions[n].token_start
-                # previous_token_end = example_word_predictions[n - 1].token_end
-                if next_tag.startswith("I-") and next_tag == current_tag.replace(
-                    "B-", "I-"
-                ):  # and previous_token_end == next_token_start:
-                    n_tags += 1
-                else:
-                    break
-
-            merged_ner_tag = example_word_predictions[i]
-            assert isinstance(
-                merged_ner_tag["tag"], str
-            ), "ERROR! merged_ner_tag.tag should be a string"
-
-            merged_ner_tag["tag"] = merged_ner_tag["tag"].split("-")[-1]
-            if n_tags > 0:
-                merged_ner_tag["char_end"] = example_word_predictions[i + n_tags][
-                    "char_end"
-                ]
-                # merged_ner_tag.token_end = example_word_predictions[i + n_tags]["token_end"]
-                assert isinstance(
-                    merged_ner_tag["char_start"], str
-                ), "ERROR! merged_ner_tag.char_start should be a string"
-                assert isinstance(
-                    merged_ner_tag["char_end"], str
-                ), "ERROR! merged_ner_tag.char_end should be a string"
-                merged_ner_tag["token"] = example[
-                    int(merged_ner_tag["char_start"]) : int(merged_ner_tag["char_end"])
-                ]
-                count["merge"] += 1 + n_tags
-            else:
-                count["replace"] += 1
-            merged_ner_tags.append(merged_ner_tag)
-        elif "-" not in current_tag:  # plain scheme
-            count["unmodified"] += 1
-            merged_ner_tag = example_word_predictions[i]
-            merged_ner_tags.append(merged_ner_tag)
-
-    assert count["merge"] + count["replace"] + count["o_tags"] + count[
-        "unmodified"
-    ] == len(
-        example_word_predictions
-    ), f"{count} -> {sum(count.values())} != {len(example_word_predictions)} | {example_word_predictions}"
-
-    if count["merge"] > 0 or count["replace"] > 0:
-        assert (
-            count["unmodified"] == 0
-        ), f"{count} -> if merge or replaced are > 0, unmodified should be == 0."
-
-    if count["unmodified"] > 0:
-        assert (
-            count["merge"] == 0
-        ), f"{count} -> if unmodified is > 0, merge should be == 0."
-        assert (
-            count["replace"] == 0
-        ), f"{count} -> if unmodified is > 0, replace should be == 0."
-
-    example_word_predictions_merged = merged_ner_tags
-    if VERBOSE:
-        print(
-            f"> merged {len(example_word_predictions_merged)} BIO-tags "
-            f"(simple replace: {count['replace']}, merge: {count['merge']}, O-tags: {count['o_tags']}, unmodified: {count['unmodified']}).\n"
-        )
-
-    return example_word_predictions_merged
-
-
-def assert_str(_object: Union[str, Dict[Any, Any]], _object_name: str) -> str:
-    assert isinstance(
-        _object, str
-    ), f"ERROR! {_object_name} = {_object} is {type(_object)} but should be string"
-    return _object
