@@ -22,10 +22,10 @@ def main(args):
         formatter.create_ner_tag_mapping_json(
             modify=args.modify
         )  # II: create ner tag mapping
-        formatter.format_data()  # III: format data
+        formatter.format_data(shuffle=True)  # III: format data
         formatter.resplit_data(val_fraction=args.val_fraction)  # IV: resplit data
-        formatter.analyze_data()  # V: analyze data
-        formatter.plot_data()  # V: analyze data
+        formatter.analyzer.analyze_data()  # V: analyze data
+        formatter.analyzer.plot_data()  # V: analyze data
 
 
 if __name__ == "__main__":
