@@ -1,7 +1,7 @@
 import pytest
 import torch
 from transformers import AutoTokenizer
-from typing import List, Dict
+from typing import List, Dict, Any
 from pkg_resources import resource_filename
 from nerblackbox.modules.ner_training.data_preprocessing.tools.csv_reader import (
     CsvReader,
@@ -12,30 +12,30 @@ from nerblackbox.modules.ner_training.data_preprocessing.tools.input_example imp
 from nerblackbox.modules.ner_training.data_preprocessing.tools.input_examples_to_tensors import (
     InputExamplesToTensors,
 )
-from nerblackbox.modules.ner_training.data_preprocessing.tools.bert_dataset import (
-    BertDataset,
+from nerblackbox.modules.ner_training.data_preprocessing.tools.encodings_dataset import (
+    EncodingsDataset,
 )
 from nerblackbox.modules.ner_training.data_preprocessing.data_preprocessor import (
     DataPreprocessor,
-    order_tag_list,
-    convert_tag_list_bio2plain,
 )
 from nerblackbox.modules.ner_training.data_preprocessing.tools.utils import (
     EncodingsKeys,
 )
 from nerblackbox.tests.utils import PseudoDefaultLogger
-from nerblackbox.modules.ner_training.ner_model import NEWLINE_TOKENS
+
+SENTENCES_ROWS_UNPRETOKENIZED = List[Dict[str, Any]]
 
 tokenizer = AutoTokenizer.from_pretrained(
     "af-ai-center/bert-base-swedish-uncased",
     do_lower_case=False,
-    additional_special_tokens=NEWLINE_TOKENS,
+    additional_special_tokens=["[newline]", "[NEWLINE]"],
     use_fast=True,
 )
 
 csv_reader = CsvReader(
     path=resource_filename("nerblackbox", "tests/test_data"),
     tokenizer=tokenizer,
+    pretokenized=True,
     do_lower_case=False,
     csv_file_separator="\t",
     default_logger=None,
@@ -56,10 +56,11 @@ class TestCsvReaderAndDataProcessor:
 
     ####################################################################################################################
     @pytest.mark.parametrize(
-        "tag_list, " "input_examples",
+        "annotation_classes, " "annotation_scheme, " "input_examples",
         [
             (
                 ["O", "PER", "ORG", "MISC"],
+                "plain",
                 {
                     "train": [
                         InputExample(
@@ -105,7 +106,8 @@ class TestCsvReaderAndDataProcessor:
     )
     def tests(
         self,
-        tag_list: List[str],
+        annotation_classes: List[str],
+        annotation_scheme: str,
         input_examples: Dict[str, List[InputExample]],
     ) -> None:
 
@@ -128,23 +130,29 @@ class TestCsvReaderAndDataProcessor:
         # 2. DataProcessor
         ##################################
         # a. get_input_examples_train
-        test_input_examples, test_tag_list = data_preprocessor.get_input_examples_train(
+        (
+            test_input_examples,
+            test_annotation,
+        ) = data_preprocessor.get_input_examples_train(
             prune_ratio={"train": 0.5, "val": 1.0, "test": 1.0},
             dataset_name=None,
         )
-        assert set(test_tag_list) == set(
-            tag_list
-        ), f"test_tag_list = {test_tag_list} != {tag_list}"
+        assert (
+            test_annotation.scheme == annotation_scheme
+        ), f"ERROR! {test_annotation.scheme} != {annotation_scheme}"
+        assert set(test_annotation.classes) == set(
+            annotation_classes
+        ), f"test_annotation.classes = {test_annotation.classes} != {annotation_classes}"
         for phase in ["train", "val", "test"]:
             assert (
                 len(test_input_examples[phase]) == 1
             ), f"ERROR! len(test_input_examples[{phase}]) = {len(test_input_examples[phase])} should be 1."
             assert (
                 test_input_examples[phase][0].text == input_examples[phase][0].text
-            ), f"phase = {phase}: test_input_examples.text = {test_input_examples[phase].text} != {input_examples[phase][0].text}"
+            ), f"phase = {phase}: test_input_examples.text = {test_input_examples[phase][0].text} != {input_examples[phase][0].text}"
             assert (
                 test_input_examples[phase][0].tags == input_examples[phase][0].tags
-            ), f"phase = {phase}: test_input_examples.tags = {test_input_examples[phase].tags} != {input_examples[phase][0].tags}"
+            ), f"phase = {phase}: test_input_examples.tags = {test_input_examples[phase][0].tags} != {input_examples[phase][0].tags}"
 
         # b. get_input_examples_predict
         test_sentences = [
@@ -171,7 +179,7 @@ class TestCsvReaderAndDataProcessor:
 
         # c. to_dataloader
         dataloader = data_preprocessor.to_dataloader(
-            input_examples, tag_list, batch_size=1
+            input_examples, annotation_classes, batch_size=1
         )
         for key in ["train", "val", "test", "predict"]:
             assert (
@@ -183,16 +191,99 @@ class TestCsvReaderAndDataProcessor:
 ########################################################################################################################
 ########################################################################################################################
 ########################################################################################################################
-class TestInputExamplesToTensorsAndBertDataset:
+class TestDataProcessor:
+    @pytest.mark.parametrize(
+        "tokens, words",
+        [
+            (
+                ["this", "is", "a", "trivial", "test"],
+                ["this", "is", "a", "trivial", "test"],
+            ),
+            (
+                ["this", "example", "contains", "hu", "##gging", "face"],
+                ["this", "example", "contains", "hugging", "face"],
+            ),
+            (
+                ["##a", "b", "##c", "##d", "##e", "##fghijkl"],
+                ["##a", "bcdefghijkl"],
+            ),
+        ],
+    )
+    def test_tokens2words(self, tokens: List[str], words: List[str]):
+        test_words = data_preprocessor._tokens2words(tokens)
+        assert test_words == words, f"test_words = {test_words} != {words}"
+
+    @pytest.mark.parametrize(
+        "data, data_pretokenized",
+        [
+            (
+                [
+                    {
+                        "text": "arbetsförmedlingen ai-center finns i stockholm.",
+                        "tags": [
+                            {
+                                "token": "arbetsförmedlingen ai-center",
+                                "tag": "ORG",
+                                "char_start": 0,
+                                "char_end": 28,
+                            },
+                            {
+                                "token": "stockholm",
+                                "tag": "LOC",
+                                "char_start": 37,
+                                "char_end": 46,
+                            },
+                        ],
+                    },
+                    {
+                        "text": "this example contains hugging face",
+                        "tags": [
+                            {
+                                "token": "hugging face",
+                                "tag": "ORG",
+                                "char_start": 22,
+                                "char_end": 34,
+                            },
+                        ],
+                    },
+                ],
+                [
+                    {
+                        "text": "arbetsförmedlingen ai - center finns i stockholm .",
+                        "tags": "B-ORG I-ORG I-ORG I-ORG O O B-LOC O",
+                    },
+                    {
+                        "text": "this example contains hugging face",
+                        "tags": "O O O B-ORG I-ORG",
+                    },
+                ],
+            ),
+        ],
+    )
+    def test_pretokenize(
+        self,
+        data: SENTENCES_ROWS_UNPRETOKENIZED,
+        data_pretokenized: List[Dict[str, str]],
+    ):
+        test_data_pretokenized = data_preprocessor._pretokenize_data(data)
+        assert (
+            test_data_pretokenized == data_pretokenized
+        ), f"test_data_pretokenized = {test_data_pretokenized} != {data_pretokenized}"
+
+
+########################################################################################################################
+########################################################################################################################
+########################################################################################################################
+class TestInputExamplesToTensorsAndEncodingsDataset:
     @pytest.mark.parametrize(
         "texts, "
         "labels, "
-        "tag_tuple, "
+        "annotation_classes, "
         "max_seq_length, "
         "true_input_ids, "
         "true_attention_mask, "
         "true_token_type_ids, "
-        "true_tag_ids, "
+        "true_labels, "
         "true_input_tokens",
         [
             # 1. single example: no truncation
@@ -225,6 +316,36 @@ class TestInputExamplesToTensorsAndBertDataset:
                     + ["[PAD]"] * 2
                 ],
             ),
+            # 1b. single example: no truncation [BIO]
+            (
+                ["arbetsförmedlingen ai-center finns i stockholm"],
+                ["B-ORG I-ORG O O B-LOC"],
+                ("O", "B-ORG", "B-LOC", "I-ORG", "I-LOC"),
+                12,
+                torch.tensor(
+                    [[101, 7093, 2842, 8126, 1011, 5410, 1121, 1045, 1305, 102, 0, 0]]
+                ),
+                torch.tensor([[1] * 10 + [0] * 2]),
+                torch.tensor([[0] * 12]),
+                torch.tensor(
+                    [[-100, 1, -100, 3, -100, -100, 0, 0, 2, -100, -100, -100]]
+                ),
+                [
+                    [
+                        "[CLS]",
+                        "arbetsförmedl",
+                        "##ingen",
+                        "ai",
+                        "-",
+                        "center",
+                        "finns",
+                        "i",
+                        "stockholm",
+                        "[SEP]",
+                    ]
+                    + ["[PAD]"] * 2
+                ],
+            ),
             # 2. single example: no truncation, [NEWLINE]
             (
                 ["arbetsförmedlingen ai-center [NEWLINE] finns i stockholm"],
@@ -370,12 +491,12 @@ class TestInputExamplesToTensorsAndBertDataset:
         self,
         texts: List[str],
         labels: List[str],
-        tag_tuple: List[str],
+        annotation_classes: List[str],
         max_seq_length: int,
         true_input_ids: torch.Tensor,
         true_attention_mask: torch.Tensor,
         true_token_type_ids: torch.Tensor,
-        true_tag_ids: torch.Tensor,
+        true_labels: torch.Tensor,
         true_input_tokens: torch.Tensor,
     ) -> None:
 
@@ -393,7 +514,7 @@ class TestInputExamplesToTensorsAndBertDataset:
         input_examples_to_tensors = InputExamplesToTensors(
             tokenizer=tokenizer,
             max_seq_length=max_seq_length,
-            tag_tuple=tuple(tag_tuple),
+            annotation_classes_tuple=tuple(annotation_classes),
             default_logger=PseudoDefaultLogger(),
         )
         encodings = input_examples_to_tensors(input_examples, predict=False)
@@ -404,7 +525,7 @@ class TestInputExamplesToTensorsAndBertDataset:
 
         for (string, true) in zip(
             EncodingsKeys,
-            [true_input_ids, true_attention_mask, true_token_type_ids, true_tag_ids],
+            [true_input_ids, true_attention_mask, true_token_type_ids, true_labels],
         ):
             assert torch.all(
                 torch.eq(encodings[string], true)
@@ -418,125 +539,24 @@ class TestInputExamplesToTensorsAndBertDataset:
             assert _test == true, f"{string} = {_test} != {true}"
 
         ##################################
-        # 2. BertDataset
+        # 2. EncodingsDataset
         ##################################
-        data = BertDataset(
+        data = EncodingsDataset(
             encodings=encodings
-        )  # data[j] = 4 torch tensors corresponding to EncodingKeys
+        )  # data[j] = Dict with 4 keys corresponding to EncodingKeys and values = torch tensors
         assert len(data) >= len(
             texts
         ), f"len(data) = {len(data)} < {len(texts)} = len(texts)"
-        for i, (string, true) in enumerate(
-            zip(
-                ["input_ids", "attention_mask", "token_type_ids", "tag_ids"],
-                [
-                    true_input_ids,
-                    true_attention_mask,
-                    true_token_type_ids,
-                    true_tag_ids,
-                ],
-            )
+        for string, true in zip(
+            ["input_ids", "attention_mask", "token_type_ids", "labels"],
+            [
+                true_input_ids,
+                true_attention_mask,
+                true_token_type_ids,
+                true_labels,
+            ],
         ):
             for j in range(len(true)):
                 assert torch.all(
-                    torch.eq(data[j][i], true[j])
-                ), f"{string} = {data[j][i]} != {true[j]}"
-
-
-########################################################################################################################
-########################################################################################################################
-########################################################################################################################
-class TestMisc:
-
-    ####################################################################################################################
-    @pytest.mark.parametrize(
-        "tag_list, " "returned_tag_list",
-        [
-            (
-                ["O", "PER", "ORG", "MISC"],
-                ["O", "PER", "ORG", "MISC"],
-            ),
-            (
-                ["O", "B-PER", "B-ORG", "B-MISC"],
-                ["O", "B-PER", "B-ORG", "B-MISC", "I-PER", "I-ORG", "I-MISC"],
-            ),
-        ],
-    )
-    def test_ensure_completeness_in_case_of_bio_tags(
-        self,
-        tag_list: List[str],
-        returned_tag_list: List[str],
-    ) -> None:
-        test_returned_tag_list = (
-            data_preprocessor._ensure_completeness_in_case_of_bio_tags(
-                tag_list=tag_list
-            )
-        )
-        assert (
-            test_returned_tag_list == returned_tag_list
-        ), f"test_returned_tag_list = {test_returned_tag_list} != {returned_tag_list}"
-
-    ####################################################################################################################
-    @pytest.mark.parametrize(
-        "tag_list, " "tag_list_ordered",
-        [
-            (
-                ["O", "PER", "ORG", "MISC"],
-                ["O", "MISC", "ORG", "PER"],
-            ),
-            (
-                ["PER", "ORG", "O", "MISC"],
-                ["O", "MISC", "ORG", "PER"],
-            ),
-            (
-                ["O", "B-PER", "I-MISC", "B-ORG", "I-PER", "B-MISC", "I-ORG"],
-                ["O", "B-MISC", "B-ORG", "B-PER", "I-MISC", "I-ORG", "I-PER"],
-            ),
-        ],
-    )
-    def test_order_tag_list(
-        self,
-        tag_list: List[str],
-        tag_list_ordered: List[str],
-    ) -> None:
-        test_tag_list_ordered = order_tag_list(tag_list)
-        assert (
-            test_tag_list_ordered == tag_list_ordered
-        ), f"test_tag_list_ordered = {test_tag_list_ordered} != {tag_list_ordered}"
-
-    ####################################################################################################################
-    @pytest.mark.parametrize(
-        "tag_list_bio, " "tag_list",
-        [
-            (
-                ["O", "B-MISC", "B-ORG", "B-PER", "I-MISC", "I-ORG", "I-PER"],
-                ["O", "MISC", "ORG", "PER"],
-            ),
-            (  # if applied to plain tag_list, nothing happens
-                ["O", "MISC", "ORG", "PER"],
-                ["O", "MISC", "ORG", "PER"],
-            ),
-        ],
-    )
-    def test_convert_tag_list_bio2plain(
-        self,
-        tag_list_bio: List[str],
-        tag_list: List[str],
-    ) -> None:
-        test_tag_list = convert_tag_list_bio2plain(tag_list_bio)
-        assert (
-            test_tag_list == tag_list
-        ), f"test_tag_list_ordered = {test_tag_list} != {tag_list}"
-
-
-if __name__ == "__main__":
-    test_csv = TestCsvReaderAndDataProcessor()
-    test_csv.tests()
-
-    test_input_examples_to_tensors = TestInputExamplesToTensorsAndBertDataset()
-    test_input_examples_to_tensors.tests()
-
-    test_misc = TestMisc()
-    test_misc.test_ensure_completeness_in_case_of_bio_tags()
-    test_misc.test_order_tag_list()
-    test_misc.test_convert_tag_list_bio2plain()
+                    torch.eq(data[j][string], true[j])
+                ), f"{string} = {data[j][string]} != {true[j]}"
