@@ -6,21 +6,20 @@ from nerblackbox.modules.ner_training.ner_model_evaluation import NerModelEvalua
 from nerblackbox.modules.ner_training.metrics.logged_metrics import LoggedMetrics
 from nerblackbox.tests.utils import PseudoDefaultLogger
 from nerblackbox.tests.test_utils import pytest_approx
+from nerblackbox.modules.ner_training.annotation_tags.annotation import Annotation
 
 
 class TestNerModelEvaluation:
 
     ner_model_evaluation = NerModelEvaluation(
         current_epoch=2,
-        tag_list=["O", "ORG", "PER"],
-        annotation_scheme="plain",
+        annotation=Annotation(["O", "ORG", "PER"]),
         default_logger=PseudoDefaultLogger(),
         logged_metrics=LoggedMetrics(),
     )
     ner_model_evaluation_bio = NerModelEvaluation(
         current_epoch=2,
-        tag_list=["O", "B-ORG", "B-PER", "I-ORG", "I-PER"],
-        annotation_scheme="bio",
+        annotation=Annotation(["O", "B-ORG", "B-PER", "I-ORG", "I-PER"]),
         default_logger=PseudoDefaultLogger(),
         logged_metrics=LoggedMetrics(),
     )
@@ -76,7 +75,7 @@ class TestNerModelEvaluation:
     )
     def test_convert_and_combine(
         self,
-        outputs: List[Union[torch.Tensor, Dict[str, Any]]],
+        outputs: List[torch.Tensor],
         np_batch: Dict[str, List[np.array]],
         np_epoch: Dict[str, Union[np.number, np.array]],
     ) -> None:
@@ -197,75 +196,124 @@ class TestNerModelEvaluation:
 
     ####################################################################################################################
     @pytest.mark.parametrize(
-        "_tags_plain, " "_tag_subset, " "_filtered_tags, " "_filtered_tags_index",
+        "_tag_subset, _tags_plain, _filtered_classes, _filtered_class_index",
         [
             (
-                {"true": np.array(["O", "ORG", "PER"]), "pred": np.array([])},
                 "all",
+                {"true": np.array(["O", "ORG", "PER"]), "pred": np.array([])},
                 ["O", "ORG", "PER"],
                 None,
             ),
             (
-                {"true": np.array(["O", "ORG", "PER"]), "pred": np.array([])},
                 "fil",
+                {"true": np.array(["O", "ORG", "PER"]), "pred": np.array([])},
                 ["ORG", "PER"],
                 None,
             ),
             (
-                {"true": np.array(["O", "ORG", "PER"]), "pred": np.array([])},
                 "O",
+                {"true": np.array(["O", "ORG", "PER"]), "pred": np.array([])},
                 ["O"],
                 None,
             ),
             (
-                {"true": np.array(["O", "ORG", "PER"]), "pred": np.array([])},
                 "PER",
+                {
+                    "true": np.array(["O", "ORG", "PER"]),
+                    "pred": np.array(["O", "PER", "ORG"]),
+                },
+                ["PER"],
+                1,
+            ),
+            (
+                "PER",
+                {
+                    "true": np.array(["O", "PER", "ORG"]),
+                    "pred": np.array(["O", "ORG", "PER"]),
+                },
                 ["PER"],
                 1,
             ),
             (
-                {"true": np.array(["O", "ORG", "PER"]), "pred": np.array([])},
                 "ORG",
+                {
+                    "true": np.array(["O", "ORG", "PER"]),
+                    "pred": np.array(["O", "PER", "ORG"]),
+                },
+                ["ORG"],
+                0,
+            ),
+            (
+                "ORG",
+                {
+                    "true": np.array(["O", "PER", "ORG"]),
+                    "pred": np.array(["O", "ORG", "PER"]),
+                },
                 ["ORG"],
                 0,
             ),
             (
-                {"true": np.array(["ORG"]), "pred": np.array([])},
                 "all",
+                {"true": np.array(["ORG"]), "pred": np.array([])},
                 ["O", "ORG", "PER"],
                 None,
             ),
             (
-                {"true": np.array(["ORG"]), "pred": np.array([])},
                 "fil",
+                {"true": np.array(["ORG"]), "pred": np.array([])},
                 ["ORG", "PER"],
                 None,
             ),
-            ({"true": np.array(["ORG"]), "pred": np.array([])}, "O", ["O"], None),
-            ({"true": np.array(["ORG"]), "pred": np.array([])}, "PER", ["PER"], None),
-            ({"true": np.array(["ORG"]), "pred": np.array([])}, "ORG", ["ORG"], 0),
-            ({"true": np.array(["PER"]), "pred": np.array([])}, "PER", ["PER"], 0),
-            ({"true": np.array(["PER"]), "pred": np.array([])}, "ORG", ["ORG"], None),
+            (
+                "O",
+                {"true": np.array(["ORG"]), "pred": np.array([])},
+                ["O"],
+                None,
+            ),
+            (
+                "PER",
+                {"true": np.array(["ORG"]), "pred": np.array([])},
+                ["PER"],
+                None,
+            ),
+            (
+                "ORG",
+                {"true": np.array(["ORG"]), "pred": np.array([])},
+                ["ORG"],
+                0,
+            ),
+            (
+                "PER",
+                {"true": np.array(["PER"]), "pred": np.array([])},
+                ["PER"],
+                0,
+            ),
+            (
+                "ORG",
+                {"true": np.array(["PER"]), "pred": np.array([])},
+                ["ORG"],
+                None,
+            ),
         ],
     )
-    def test_get_filtered_tags(
+    def test_get_filtered_classes(
         self,
-        _tags_plain: Dict[str, np.array],
         _tag_subset: str,
-        _filtered_tags: List[str],
-        _filtered_tags_index: Optional[int],
+        _tags_plain: Dict[str, np.array],
+        _filtered_classes: List[str],
+        _filtered_class_index: Optional[int],
     ) -> None:
         (
-            test_filtered_tags,
-            test_filtered_tags_index,
-        ) = self.ner_model_evaluation._get_filtered_tags(_tag_subset, _tags_plain)
+            test_filtered_classes,
+            test_filtered_class_index,
+        ) = self.ner_model_evaluation._get_filtered_classes(_tag_subset, _tags_plain)
 
         assert (
-            test_filtered_tags == _filtered_tags
-        ), f"test_filtered_tags = {test_filtered_tags} != {_filtered_tags}"
+            test_filtered_classes == _filtered_classes
+        ), f"test_filtered_classes = {test_filtered_classes} != {_filtered_classes}"
         assert (
-            test_filtered_tags_index == _filtered_tags_index
-        ), f"test_filtered_tags_index = {test_filtered_tags_index} != {_filtered_tags_index}"
+            test_filtered_class_index == _filtered_class_index
+        ), f"test_filtered_class_index = {test_filtered_class_index} != {_filtered_class_index}"
 
     ####################################################################################################################
     @pytest.mark.parametrize(
@@ -287,6 +335,10 @@ class TestNerModelEvaluation:
                     "entity_fil_precision_macro": 1.0,
                     "entity_fil_f1_micro": 1.0,
                     "entity_fil_f1_macro": 1.0,
+                    "entity_fil_asr_abidance": 1.0,
+                    "entity_fil_asr_precision_micro": 1.0,
+                    "entity_fil_asr_recall_micro": 1.0,
+                    "entity_fil_asr_f1_micro": 1.0,
                     "token_fil_numberofclasses_macro": 2,
                     "token_fil_recall_micro": 1.0,
                     "token_fil_recall_macro": 1.0,
@@ -444,6 +496,10 @@ class TestNerModelEvaluation:
                     "entity_fil_recall_macro": 0.0,
                     "entity_fil_f1_micro": 0.0,
                     "entity_fil_f1_macro": 0.0,
+                    "entity_fil_asr_abidance": 1.0,
+                    "entity_fil_asr_precision_micro": 0.0,
+                    "entity_fil_asr_recall_micro": 0.0,
+                    "entity_fil_asr_f1_micro": 0.0,
                     "entity_PER_precision": -1.0,
                     "entity_PER_recall": 0.0,
                     "entity_PER_f1": -1.0,
@@ -453,18 +509,18 @@ class TestNerModelEvaluation:
                     "token_all_loss": 2.0,
                     "token_all_acc": 0.75,
                     "token_all_precision_micro": 0.75,
-                    "token_all_precision_macro": -1.0,
+                    "token_all_precision_macro": 0.50,
                     "token_all_recall_micro": 0.75,
                     "token_all_recall_macro": 0.66,
                     "token_all_f1_micro": 0.75,
-                    "token_all_f1_macro": -1.0,
+                    "token_all_f1_macro": 0.55,
                     "token_fil_numberofclasses_macro": 1,
                     "token_fil_precision_micro": 0.5,
-                    "token_fil_precision_macro": -1.0,
+                    "token_fil_precision_macro": 0.25,
                     "token_fil_recall_micro": 0.5,
                     "token_fil_recall_macro": 0.5,
                     "token_fil_f1_micro": 0.5,
-                    "token_fil_f1_macro": -1.0,
+                    "token_fil_f1_macro": 0.33,
                     "token_O_precision": 1.0,
                     "token_O_recall": 1.0,
                     "token_O_f1": 1.0,
@@ -507,6 +563,10 @@ class TestNerModelEvaluation:
                     "entity_fil_recall_macro": 0.0,
                     "entity_fil_f1_micro": 0.0,
                     "entity_fil_f1_macro": 0.0,
+                    "entity_fil_asr_abidance": 1.0,
+                    "entity_fil_asr_precision_micro": 0.0,
+                    "entity_fil_asr_recall_micro": 0.0,
+                    "entity_fil_asr_f1_micro": 0.0,
                     "entity_PER_precision": -1.0,
                     "entity_PER_recall": 0.0,
                     "entity_PER_f1": -1.0,
@@ -516,18 +576,18 @@ class TestNerModelEvaluation:
                     "token_all_loss": 2.0,
                     "token_all_acc": 0.75,
                     "token_all_precision_micro": 0.75,
-                    "token_all_precision_macro": -1.0,
+                    "token_all_precision_macro": 0.5,
                     "token_all_recall_micro": 0.75,
                     "token_all_recall_macro": 0.66,
                     "token_all_f1_micro": 0.75,
-                    "token_all_f1_macro": -1.0,
+                    "token_all_f1_macro": 0.55,
                     "token_fil_numberofclasses_macro": 1,
                     "token_fil_precision_micro": 0.5,
-                    "token_fil_precision_macro": -1.0,
+                    "token_fil_precision_macro": 0.25,
                     "token_fil_recall_micro": 0.5,
                     "token_fil_recall_macro": 0.5,
                     "token_fil_f1_micro": 0.5,
-                    "token_fil_f1_macro": -1.0,
+                    "token_fil_f1_macro": 0.33,
                     "token_O_precision": 1.0,
                     "token_O_recall": 1.0,
                     "token_O_f1": 1.0,
@@ -613,6 +673,10 @@ class TestNerModelEvaluation:
                     "entity_fil_recall_macro": 0.0,
                     "entity_fil_f1_micro": 0.0,
                     "entity_fil_f1_macro": 0.0,
+                    "entity_fil_asr_abidance": 1.0,
+                    "entity_fil_asr_precision_micro": 0.0,
+                    "entity_fil_asr_recall_micro": 0.0,
+                    "entity_fil_asr_f1_micro": 0.0,
                     "entity_PER_precision": -1.0,
                     "entity_PER_recall": 0.0,
                     "entity_PER_f1": -1.0,
@@ -622,18 +686,18 @@ class TestNerModelEvaluation:
                     "token_all_loss": 2.0,
                     "token_all_acc": 0.75,
                     "token_all_precision_micro": 0.75,
-                    "token_all_precision_macro": -1.0,
+                    "token_all_precision_macro": 0.50,
                     "token_all_recall_micro": 0.75,
                     "token_all_recall_macro": 0.66,
                     "token_all_f1_micro": 0.75,
-                    "token_all_f1_macro": -1.0,
+                    "token_all_f1_macro": 0.55,
                     "token_fil_numberofclasses_macro": 1,
                     "token_fil_precision_micro": 0.5,
-                    "token_fil_precision_macro": -1.0,
+                    "token_fil_precision_macro": 0.25,
                     "token_fil_recall_micro": 0.5,
                     "token_fil_recall_macro": 0.5,
                     "token_fil_f1_micro": 0.5,
-                    "token_fil_f1_macro": -1.0,
+                    "token_fil_f1_macro": 0.33,
                     "token_O_precision": 1.0,
                     "token_O_recall": 1.0,
                     "token_O_f1": 1.0,
@@ -651,7 +715,7 @@ class TestNerModelEvaluation:
     def test_execute(
         self,
         phase: str,
-        outputs: List[Union[torch.Tensor, Dict[str, Any]]],
+        outputs: List[torch.Tensor],
         epoch_metrics: Dict[str, np.array],
         epoch_loss: float,
     ) -> None:
@@ -670,15 +734,3 @@ class TestNerModelEvaluation:
             assert test_epoch_metrics[k] == pytest_approx(
                 epoch_metrics[k]
             ), f"key = {k}: test_epoch_metrics = {test_epoch_metrics[k]} != {epoch_metrics[k]}"
-
-
-if __name__ == "__main__":
-    test = TestNerModelEvaluation()
-    test.test_convert_and_combine()
-    test.test_reduce_and_flatten()
-    test.test_convert_tag_ids_to_tags()
-    test.test_get_rid_of_special_tag_occurrences()
-    test.test_get_filtered_tags()
-    test.test_compute_metrics_for_tags_subset()
-    test.test_compute_metrics()
-    test.test_execute()
